from typing import List, Optional, Union, Iterator, Tuple

from musicscore import Chord
from musicscore.finalize import FinalizeMixin
from musicscore.quantize import QuantizeMixin
from musicxml.xmlelement.xmlelement import XMLPart, XMLScorePart

from musicscore.exceptions import IdHasAlreadyParentOfSameTypeError, IdWithSameValueExistsError, VoiceIsFullError, \
    AlreadyFinalizedError
from musicscore.measure import Measure
from musicscore.musictree import MusicTree
from musicscore.time import Time
from musicscore.xmlwrapper import XMLWrapper

__all__ = ['Id', 'ScorePart', 'Part']


class Id:
    __refs__ = []

    def __init__(self, value):
        self._parents = []
        self._value = None
        self.value = value
        self.__refs__.append(self)

    @classmethod
    def _check_value(cls, val):
        for obj in cls.__refs__:
            if obj.value == val:
                raise IdWithSameValueExistsError

    @property
    def value(self) -> str:
        """
        - val: a unique id. If not unique IdWithSameValueExistsError is raised.
        - All parents ids will be updated.
        """
        return self._value

    @value.setter
    def value(self, val):
        self._check_value(val)
        self._value = val
        for parent in self.get_parents():
            self.update_parents_id(parent)

    def delete(self) -> None:
        """
        Removes Id instance from class attribute __refs__ before deleting.
        """
        if self in self.__refs__:
            self.__refs__.remove(self)
        del self

    def update_parents_id(self, parent: XMLWrapper) -> None:
        """
        Sets parent's xml_object.id to self.value
        """
        parent.xml_object.id = self.value

    def add_parent(self, obj: XMLWrapper) -> None:
        """
        Gets object to Id as parent. Parents id gets updated.
        """
        if obj.__class__ in [type(parent) for parent in self.get_parents()]:
            raise IdHasAlreadyParentOfSameTypeError()
        self._parents.append(obj)
        self.update_parents_id(obj)

    def get_parents(self) -> List[XMLWrapper]:
        """
        Gets Id's parent objects
        """
        return self._parents

    def __repr__(self):
        return f"{self.__class__}:{self.value} at {id(self)}"

    def __del__(self):
        if self in self.__refs__:
            self.__refs__.remove(self)


class ScorePart(XMLWrapper):
    _ATTRIBUTES = {'part'}

    XMLClass = XMLScorePart

    def __init__(self, part, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._part = None
        self.part = part

    @property
    def part(self) -> 'Part':
        """
        val type: :obj:~`Part`
        obj:~`musicxml.xmlelement.xmlelement.XMLPartName`will be updated
        :obj:~`musicscore.part.Id` is set or updated

        :return: :obj:~`Part`
        """
        return self._part

    @part.setter
    def part(self, val):
        if not isinstance(val, Part):
            raise TypeError
        self._part = val
        if self in self.part.id_.get_parents():
            self.part.id_.update_parents_id(self)
        else:
            self.part.id_.add_parent(self)
        self._update_name()

    def _update_abbreviation(self):
        self.xml_object.xml_part_abbreviation = self.part.abbreviation

    def _update_name(self):
        self.xml_object.xml_part_name = self.part.name


class Part(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    _ATTRIBUTES = {'id_', 'name', 'abbreviation'}
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)
    XMLClass = XMLPart

    def __init__(self, id, name=None, abbreviation=None, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._id = None
        self.id_ = id
        self._score_part = ScorePart(part=self)
        self._name = None
        self.name = name
        self._abbreviation = None
        self.abbreviation = abbreviation
        self._current_measures = {}

    def _add_to_next_measure(self, current_measure, chord, staff_number, voice_number):
        if current_measure.next:
            current_measure = current_measure.next
        else:
            current_measure = self.add_measure()
        current_measure._add_chord(chord, staff_number=staff_number, voice_number=voice_number)
        return current_measure

    def _set_first_current_measure(self, staff_number, voice_number):
        for m in self.get_children():
            if m.get_voice(staff_number=staff_number, voice_number=voice_number):
                self.set_current_measure(staff_number, voice_number, m)
                return m

    @property
    def abbreviation(self) -> Optional[str]:
        return self._abbreviation

    @abbreviation.setter
    def abbreviation(self, val):
        self._abbreviation = val
        try:
            self.score_part._update_abbreviation()
        except AttributeError:
            pass

    @property
    def id_(self) -> Id:
        """
        :rtype: :obj:`~musicscore.part.Id`, str
        :return: :obj:`~musicscore.part.Id`
        """
        return self._id

    @id_.setter
    def id_(self, val):
        if isinstance(val, Id):
            self._id = val
        elif isinstance(self._id, Id):
            self._id.value = val
        else:
            self._id = Id(val)
        if self in self.id_.get_parents():
            self.id_.update_parents_id(self)
        else:
            self.id_.add_parent(self)

    @property
    def name(self) -> str:
        """
        Set and get name. Setting tries toupdate name of :obj:`musicscore.part.score_part`

        :type: str
        :return: part's name. If no name is set part's :obj:`id_` is returned.
        """
        if self._name is not None:
            return self._name
        else:
            return ''

    @name.setter
    def name(self, val):
        self._name = val
        try:
            self.score_part._update_name()
        except AttributeError:
            pass

    @property
    def score_part(self) -> ScorePart:
        """
        :return: the  :obj:`~musicscore.part.ScorePart` which is associated with this part.
        """
        return self._score_part

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def add_child(self, child: Measure) -> Measure:
        """
        Check and add child to list of children. Child's parent is set to self.

        :param child: :obj:`~musicscore.measure.Measure`
        :return: child
        :rtype: :obj:`~musicscore.measure.Measure`

        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_child')
        super().add_child(child)
        self.xml_object.add_child(child.xml_object)
        return child

    def add_chord(self, chord: 'Chord', *, staff_number: Optional[int] = None, voice_number: Optional[int] = 1) -> None:
        """
        - Adds a chord to the specified voice in current measure (see :obj:`get_current_measure()`).
        - If no current measure is set the first measure is selected.
        - If part has still no measures one measure is added.
        - If the specified voice in current measure is full chord is added to the voice with the same number in the next measure. If no
          next measure exists one measure is added
        - If a leftover chord remains after adding chord, it is added to voice's :obj:`~musicscore.voice.Voice.leftover_chord` and is
          added to so many next measures as needed.

        :param chord: obj:`~musicscore.chord.Chord` required
        :param staff_number: positive int, None. If None is set to 1.
        :param voice_number: positive_int
        :return: None
        """
        if not isinstance(chord, Chord):
            raise TypeError(f'{chord} must be of type Chord.')
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_chord')

        for gch in chord._grace_chords['before']:
            self.add_chord(gch, staff_number=staff_number, voice_number=voice_number)

        if staff_number is None:
            staff_number = 1
        current_measure = self.get_current_measure(staff_number=staff_number, voice_number=voice_number)

        if not current_measure:
            if self.get_children():
                current_measure = self.get_children()[0]
            else:
                current_measure = self.add_measure()
        try:
            current_measure._add_chord(chord, staff_number=staff_number, voice_number=voice_number)
        except VoiceIsFullError:
            current_measure = self._add_to_next_measure(current_measure, chord, staff_number, voice_number)

        for gch in chord._grace_chords['after']:
            self.add_chord(gch, staff_number=staff_number, voice_number=voice_number)

        leftover_chord = current_measure.get_voice(staff_number=staff_number, voice_number=voice_number).leftover_chord
        while leftover_chord:
            current_measure = self._add_to_next_measure(current_measure, leftover_chord, staff_number,
                                                        voice_number)
            leftover_chord = current_measure.get_voice(staff_number=staff_number,
                                                       voice_number=voice_number).leftover_chord

    def add_measure(self, time: Optional[Union[Time, List, Tuple]] = None, number: Optional[int] = None) -> Measure:
        """
        - Creates and adds a :obj:`~musicscore.measure.Measure` to part.
        - If time is not given last measure's :obj:`~musicscore.time.Time` is copied. Its :obj:`~musicscore.time.Time.show` to
          ``False``
        - If number is not given last measure's number is incremented.
        - New measure's :obj:`~musicscore.key.Key` is a copy of last measure's :obj:`~musicscore.key.Key`. Its
          :obj:`~musicscore.key.Key.show` is set to ``False``

        :param time: Time, (numerator, denominator), None
        :param number: positive int, None
        :return: created and added :obj:`~musicscore.measure.Measure`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_measure')
        previous_measure = self.get_children()[-1] if self.get_children() else None
        if not time:
            if previous_measure:
                time = previous_measure.time.__copy__()
                time.show = False
            else:
                time = Time(4, 4)
        else:
            if not isinstance(time, Time):
                time = Time(*time)

        if not number:
            if previous_measure:
                number = previous_measure.number + 1
            else:
                number = 1

        m = Measure(number=number, time=time)
        child = self.add_child(m)
        if previous_measure:
            m.key = previous_measure.key.__copy__()
            m.key.show = False

            for staff in previous_measure.get_children():
                st = m.add_staff(staff_number=staff.number)
                if st.clef:
                    st.clef = staff.clef.__copy__()
                    st.clef.show = False
                if st:
                    st.add_voice(voice_number=1)
        else:
            m.add_voice(staff_number=None, voice_number=1)

        return child

    def get_current_measure(self, staff_number: Optional[int] = 1, voice_number: int = 1):
        """
        Gets current measure for adding :obj:`~musicscore.chord.Chord` to a specific :obj:`~musicscore.voice.Voice`

        staff_number None is set to 1

        :param staff_number: positive int, None
        :param voice_number: positive int
        """
        if staff_number is None:
            staff_number = 1
        try:
            return self._current_measures[staff_number][voice_number]
        except KeyError:
            return self._set_first_current_measure(staff_number=staff_number, voice_number=voice_number)

    def get_children(self) -> List['Measure']:
        return super().get_children()

    def get_parent(self) -> 'Score':
        """
        :return: parent
        :rtype: :obj:`~musicscore.score.Score`
        """
        return super().get_parent()

    def set_current_measure(self, staff_number: int, voice_number: int, measure: Measure) -> None:
        """
        Sets current measure for adding :obj:`~musicscore.chord.Chord` to a specific :obj:`~musicscore.voice.Voice`

        :param staff_number: positive int
        :param voice_number: positive int
        :param measure: :obj:`musicscore.measure.Measure`
        :return: None
        """
        if staff_number is None:
            staff_number = 1
        if not isinstance(measure, Measure):
            raise TypeError(f"{measure} must be of type 'Measure'.")
        if self._current_measures.get(staff_number):
            self._current_measures[staff_number][voice_number] = measure
        else:
            self._current_measures[staff_number] = {voice_number: measure}



import cProfile
import timeit
from contextlib import redirect_stdout
from pathlib import Path

from musicxml.xmlelement.xmlelement import XMLLeftMargin


def hello_xml_left_margin():
    XMLLeftMargin(10)


def time_hello_xml_left_margin():
    print(timeit.timeit(hello_xml_left_margin, number=1000))


with open(str(Path(__file__)) + 'xml_left_margin.txt', '+w') as f:
    with redirect_stdout(f):
        cProfile.run('time_hello_xml_left_margin()', sort="tottime")



from pathlib import Path

from musicscore import Time
from musicscore.chord import Chord
from musicscore.score import Score
from musicscore.tests.util import IdTestCase, generate_all_triplets, generate_all_quintuplets, generate_all_sextuplets, \
    generate_all_septuplets


class TestTuplets2(IdTestCase):
    def test_quarter_tuplets(self):
        score = Score()
        part = score.add_part('p1')
        t = Time(4, 4)
        t.actual_signatures = [1, 2, 1, 2]
        part.add_measure(t)
        tuplets = generate_all_triplets() + generate_all_quintuplets()


        tuplets = [x * 2 for t in tuplets for x in t]

        for t in tuplets:
            part.add_chord(Chord(60, t))

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



"""
Different Arpeggio directions (normal, up, down, non-arpeggiate)
"""
from pathlib import Path

from musicscore import Score, Chord, C, E, G
from musicscore.tests.util import IdTestCase
from musicxml import XMLArpeggiate


class TestLily32d(IdTestCase):
    def test_lily_32d_Arpeggio(self):
        score = Score()
        part = score.add_part('p1')

        types = ['normal', 'up', 'normal', 'down', 'normal', 'none', 'normal']
        chords = [Chord([C(4), E(5), G(5)], 1) for _ in types]

        for t, ch in zip(types, chords):
            ch.arpeggio = t

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path
from unittest import TestCase

from musicscore import SimpleFormat
from musicscore.accidental import Accidental
from musicscore.chord import Chord
from musicscore.measure import generate_measures
from musicscore.midi import Midi
from musicscore.part import Part
from musicscore.score import Score
from musicscore.time import Time


class TestHelloChromatics(TestCase):
    def test_chromatic_quarter_tone(self):
        """
        Up and down chromatic scales with quarter tones
        """
        s = Score()
        p = s.add_child(Part('p1'))
        """
        Using SimpleFormat:
        """
        midi_values = [val / 2 for val in range(120, 120 + 48 + 1)]
        midi_values = midi_values + list(reversed(midi_values[:-1]))
        print(midi_values)
        sf = SimpleFormat(midis=midi_values)
        for chord in sf.chords:
            p.add_chord(chord)
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)

        # measures = generate_measures(times=[Time(qd, 4) for qd in quarter_durations])
        # for m in measures:
        #     p.add_child(m)
        # """
        # Adding chords
        # """
        # min_midi = 60
        # for m, qd in zip(measures, quarter_durations):
        #     m._add_chord(Chord(quarter_duration=qd, midis=Midi(min_midi + 10 - qd, accidental=Accidental(mode='sharp'))))
        # """
        # Exporting
        # """
        # xml_path = Path(__file__).with_suffix('.xml')
        # s.export_xml(xml_path)



from unittest import TestCase

from musicscore.beat import Beat
from musicscore.chord import Chord
from musicscore.exceptions import AddChordError, VoiceIsFullError
from musicscore.measure import Measure
from musicscore.quarterduration import QuarterDuration
from musicscore.staff import Staff
from musicscore.voice import Voice


def create_voice():
    voice = Voice()
    st = Staff()
    m = Measure(1)
    m.add_child(st)
    st.add_child(voice)
    return voice


class TestBeat(TestCase):

    def test_beat_quarter_duration(self):
        b = Beat()
        assert b.quarter_duration == 1
        b.quarter_duration = 4
        assert b.quarter_duration == 4

    def test_beat_previous(self):
        b = Beat()
        assert b.previous is None
        v = create_voice()
        v.update_beats(1 / 4, 1 / 8, 1 / 2, 1)
        assert v.get_children()[0].previous is None
        assert v.get_children()[1].previous.quarter_duration == 1 / 4
        assert v.get_children()[2].previous.quarter_duration == 1 / 8
        assert v.get_children()[3].previous.quarter_duration == 1 / 2

    def test_beat_next(self):
        b = Beat()
        assert b.next is None
        v = create_voice()

        v.update_beats(1 / 4, 1 / 8, 1 / 2, 1)
        assert v.get_children()[0].next.quarter_duration == 1 / 8
        assert v.get_children()[1].next.quarter_duration == 1 / 2
        assert v.get_children()[2].next.quarter_duration == 1
        assert v.get_children()[3].next is None

    def test_beat_offset(self):
        b = Beat()
        assert b.offset is None
        v = create_voice()
        v.update_beats(1 / 4, 1 / 8, 1 / 2, 1)
        assert v.get_children()[0].offset == 0
        assert v.get_children()[1].offset == 1 / 4
        assert v.get_children()[2].offset == 1 / 4 + 1 / 8
        assert v.get_children()[3].offset == 1 / 4 + 1 / 8 + 1 / 2


class TestBeatAddChild(TestCase):
    def test_beat_add_child(self):
        v = create_voice()
        beats = v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        assert v.get_current_beat() == beats[0]
        chord1 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 8))[0]
        assert chord1.up == beats[0]
        assert v.get_current_beat() == beats[0]
        chord2 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 8))[0]
        assert chord2.up == beats[0]
        assert v.get_current_beat() == beats[1]
        chord3 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 4))[0]
        assert chord3.up == beats[1]
        assert v.get_current_beat() == beats[2]
        chord4 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 2))[0]
        assert chord4.up == beats[2]
        with self.assertRaises(VoiceIsFullError):
            v.get_current_beat()
        assert beats[3].is_filled
        assert v.leftover_chord is None

        v = create_voice()
        beats = v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        assert v.get_current_beat() == beats[0]
        v.get_current_beat().add_child(Chord(midis=60, quarter_duration=1 / 8))
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=1 / 4))
        assert v.get_current_beat() == beats[1]
        assert [ch.quarter_duration for ch in beats[0].get_children()] == [1 / 8, 1 / 8]
        assert [ch.midis[0].value for ch in beats[0].get_children()] == [60, 61]
        assert [ch.quarter_duration for ch in beats[1].get_children()] == [1 / 8]
        assert [ch.midis[0].value for ch in beats[1].get_children()] == [61]

    def test_add_child_with_fractional_quarter_duration_1(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=0.5))
        v.get_current_beat().add_child(Chord(midis=62, quarter_duration=3.5))
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [0.5, 0.5]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61, 62]
                assert beat.is_filled
            elif index == 1:
                assert [ch.quarter_duration for ch in beat.get_children()] == [3]
                assert [ch.midis[0].value for ch in beat.get_children()] == [62]
                assert beat.is_filled
            else:
                assert beat.get_children() == []
                assert beat.is_filled

    def test_add_child_with_fractional_quarter_duration_2(self):
        v = create_voice()
        beats = v.update_beats(1, 1.5, 1)
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=QuarterDuration(2, 5)))
        v.get_current_beat().add_child(Chord(midis=62, quarter_duration=4))
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [QuarterDuration(2, 5),
                                                                               QuarterDuration(3, 5)]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61, 62]
                assert beat.is_filled
            elif index == 1:
                assert [ch.quarter_duration for ch in beat.get_children()] == [1.5]
                assert [ch.midis[0].value for ch in beat.get_children()] == [62]
                assert beat.is_filled
            elif index == 2:
                assert [ch.quarter_duration for ch in beat.get_children()] == [1]
                assert [ch.midis[0].value for ch in beat.get_children()] == [62]
                assert beat.is_filled
        assert v.leftover_chord.quarter_duration == 4 - 2.5 - 0.6


class TestBeatSplitChord(TestCase):
    def test_split_chord_different_accidental_instances(self):
        v1 = create_voice()
        v1.update_beats(1, 1.5)
        chord = Chord(midis=61, quarter_duration=2.5)
        v1.get_current_beat().add_child(chord)
        ch1, ch2 = v1.get_chords()

        assert ch1.midis[0].accidental != ch2.midis[0].accidental

    def test_split_chord_accidentals(self):
        v1 = create_voice()
        v1.update_beats(1, 1.5, 1)
        chord = Chord(midis=61, quarter_duration=7)
        chord.midis[0].accidental.show = True
        assert chord.midis[0].accidental.xml_object.value_ == 'sharp'
        v1.get_current_beat().add_child(chord)
        v2 = create_voice()
        v2.update_beats(1, 1, 1)
        v2.get_current_beat().add_child(v1.leftover_chord)
        all_chords = v1.get_chords() + v2.get_chords()
        v1.up.up._update_divisions()
        v2.up.up._update_divisions()
        for b in v1.get_children() + v2.get_children():
            b.finalize()

        assert [ch.midis[0].accidental.xml_object.value_ if ch.midis[0].accidental.xml_object else None for ch in
                all_chords] == ['sharp',
                                None,
                                None,
                                None]
        all_chords[1].midis[0].accidental.show = True

        assert [ch.midis[0].accidental.xml_object.value_ if ch.midis[0].accidental.xml_object else None for ch in
                all_chords] == ['sharp',
                                'sharp',
                                None,
                                None]

    def test_split_not_writable_chords(self):
        v = create_voice()
        v.update_beats(1)
        v._add_chord(Chord(60, 5 / 6))
        v._add_chord(Chord(60, 1 / 6))
        assert [ch.quarter_duration for ch in v.get_chords()] == [5 / 6, 1 / 6]
        v.get_beat(1)._split_not_writable_chords()
        assert [ch.quarter_duration for ch in v.get_chords()] == [1 / 2, 1 / 3, 1 / 6]

    def test_add_child_5_leftover(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        split_children = v.get_current_beat().add_child(Chord(midis=61, quarter_duration=6))
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [4]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61]
                assert beat.is_filled
            else:
                assert beat.get_children() == []
                assert beat.is_filled
        assert split_children[0].midis[0].is_tied_to_next
        assert v.leftover_chord.quarter_duration == 2
        assert not v.leftover_chord.split
        assert v.leftover_chord.midis[0].value == 61
        assert v.leftover_chord.midis[0].is_tied_to_previous

    def test_beat_add_child_1_3_no_split(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        v.get_current_beat().add_child(Chord(midis=60, quarter_duration=1))
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=3))
        assert v.leftover_chord is None
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [1]
                assert [ch.midis[0].value for ch in beat.get_children()] == [60]
                assert beat.is_filled
            elif index == 1:
                assert [ch.quarter_duration for ch in beat.get_children()] == [3]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61]
                assert beat.is_filled
            else:
                assert beat.get_children() == []
                assert beat.is_filled

    def test_beat_add_child_4_no_split(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        chord = v.get_current_beat().add_child(Chord(60, quarter_duration=4))
        assert v.leftover_chord is None
        for index, beat in enumerate(beats):
            if index == 0:
                assert beat.get_children() == chord
                assert beat.get_children()[0].quarter_duration == 4
                assert beat.is_filled
            else:
                assert beat.is_filled

    def test_split_tied_chords(self):
        v = create_voice()
        v.update_beats(1, 1, 1, 1)
        ch1 = Chord(60, quarter_duration=1.5)
        ch1.add_tie('start')
        split_chords_1 = v.get_current_beat().add_child(ch1)
        assert split_chords_1[-1].midis[0].is_tied_to_next

        ch2 = Chord([60, 63], quarter_duration=2)
        ch2.midis[1].add_tie('start')
        split_chords_2 = v.get_current_beat().add_child(ch2)
        assert not split_chords_2[-1].midis[0].is_tied_to_next
        assert split_chords_2[-1].midis[1].is_tied_to_next

    def test_add_chord_exception(self):
        b = Beat()
        with self.assertRaises(AddChordError):
            b.add_chord()



from typing import Optional, List

from musicscore.musictree import MusicTree
from musicscore.exceptions import NoteTypeError, NoteHasNoParentChordError, NoteMidiHasNoParentChordError
from musicscore.midi import Midi
from musicscore.quarterduration import QuarterDurationMixin
from musicscore.util import note_types
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLNote, XMLDot, XMLGrace, XMLRest, XMLTie, XMLNotations, XMLTied

__all__ = ['Note', 'tie', 'untie']


def tie(*notes):
    notes[0].start_tie()
    if len(notes) > 1:
        for note in notes[1:-1]:
            note.stop_tie()
            note.start_tie()
        notes[-1].stop_tie()


def untie(*notes):
    notes[0].remove_tie('start')
    if len(notes) > 1:
        for note in notes[1:-1]:
            note.remove_tie('stop')
            note.remove_tie('start')
        notes[-1].remove_tie('stop')


class Note(MusicTree, XMLWrapper, QuarterDurationMixin):
    _ATTRIBUTES = {'midi', 'quarter_duration', 'parent_chord', 'number_of_dots', 'is_tied', 'is_tied_to_previous'}

    XMLClass = XMLNote

    def __init__(self, midi, quarter_duration=None, *args, **kwargs):
        self._midi = None
        self._parent_chord = midi.parent_chord
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._type = None
        self._number_of_dots = None

        super().__init__(quarter_duration=quarter_duration)
        self.midi = midi
        self._parent = self.parent_chord
        self._update_xml_notehead()
        self._update_xml_voice()
        self._update_xml_staff()

    @staticmethod
    def _check_xml_duration_value(duration):
        if int(duration) != duration:
            raise ValueError(f'product of quarter_duration and divisions {duration} must be an integer')
        if duration < 0:
            raise ValueError

    def _add_child(self, child: Midi) -> Midi:
        """
        Check and add child to list of children. Child's parent is set to self.

        :param child: :obj:`~musicscore.midi.Midi`
        :return: child
        :rtype: :obj:`~musicscore.midi.Midi`
        """
        return super().add_child(child)

    def _set_quarter_duration(self, val):
        old_quarter_duration = self._quarter_duration
        super()._set_quarter_duration(val)
        if self._type is None and self._quarter_duration != 0:
            try:
                note_types[self._quarter_duration.as_integer_ratio()]
            except KeyError:
                msg = f"A note with quarter_duration {self._quarter_duration} and offset {self.up.offset} is not writable and must be " \
                      f"split."
                self._quarter_duration = old_quarter_duration
                raise NoteTypeError(msg)

    def _set_xml_tied(self, val):
        if not self.xml_notations:
            self.xml_notations = XMLNotations()
        tied_xml_objects = self.xml_notations.find_children('XMLTied')
        tied_xml_types = [t.type for t in tied_xml_objects]
        if val in tied_xml_types:
            pass
        elif val == 'stop' and 'start' in tied_xml_types:
            tied_xml_objects[0].type = 'stop'
            self.xml_notations.add_child(XMLTied(type='start'))
        else:
            self.xml_notations.add_child(XMLTied(type=val))

    def _update_ties(self):
        if 'stop' in self.midi._ties:
            self.stop_tie()
        else:
            self.remove_tie('stop')
        if 'start' in self.midi._ties:
            self.start_tie()
        else:
            self.remove_tie('start')

    def _update_xml_accidental(self):
        self.xml_object.xml_accidental = self.midi.accidental.xml_object

    def _update_xml_dots(self, number_of_dots):
        dots = self.xml_object.find_children('XMLDot')
        if number_of_dots > len(dots):
            diff = number_of_dots - len(dots)
            while diff:
                self.xml_object.add_child(XMLDot())
                diff -= 1
        elif number_of_dots < len(dots):
            for dot in dots[number_of_dots:]:
                dot.get_parent().remove(dot)
        else:
            pass

    def _update_xml_duration(self):
        duration = float(self.quarter_duration) * self.get_parent_measure().get_divisions()
        self._check_xml_duration_value(duration)
        duration = int(duration)
        if duration == 0:
            if self.midi and self.midi.value == 0:
                raise ValueError('A rest cannot be a grace note.')
            self.xml_object.xml_duration = None
            if not self.xml_object.xml_grace:
                self.xml_object.xml_grace = XMLGrace()
        else:
            self.xml_object.xml_grace = None
            self.xml_object.xml_duration = duration

    def _update_xml_pitch_or_rest(self):
        if self.midi.value == 0 and self.quarter_duration == 0:
            raise ValueError('A rest cannot be a grace note.')
        pitch_or_rest = self.midi.get_pitch_or_rest()
        if isinstance(pitch_or_rest, XMLRest):
            if self.xml_object.xml_pitch:
                self.xml_object.xml_pitch = None
            self.xml_object.xml_rest = pitch_or_rest
            self.xml_object.xml_notehead = None
        else:
            if self.xml_object.xml_rest:
                self.xml_object.xml_rest = None
            self.xml_object.xml_pitch = pitch_or_rest

    def _update_xml_staff(self):
        self.xml_object.xml_staff = self.get_staff_number()

    def _update_xml_type(self):
        if self._type is None:
            if self.quarter_duration != 0:
                self.xml_type = self.quarter_duration.get_type()
            else:
                self.xml_type = None

    def _update_xml_voice(self):
        self.xml_object.xml_voice = str(self.get_voice_number())

    def _update_xml_notations(self):
        """
        If ``self.xml_object.xml_notations`` has children of types :obj:`~musicxml.xmlelement.xmlelement.XMLArticulation` oder
        :obj:`~musicxml.xmlelement.xmlelement.XMLTechnical`, :obj:`~musicxml.xmlelement.xmlelement.XMLOrnaments`,
        :obj:`~musicxml.xmlelement.xmlelement.XMLDynamics`
        which have no children themselves, these will be removed.
        ``self.xml_object.xml_notations`` will be removed itself if it has no children.

        :return: None
        """
        if self.xml_object.xml_notations:
            if self.xml_object.xml_notations.xml_articulations and not self.xml_object.xml_notations.xml_articulations.get_children():
                self.xml_object.xml_notations.remove(self.xml_object.xml_notations.xml_articulations)

            if self.xml_object.xml_notations.xml_technical and not self.xml_object.xml_notations.xml_technical.get_children():
                self.xml_object.xml_notations.remove(self.xml_object.xml_notations.xml_technical)

            if self.xml_object.xml_notations.xml_ornaments and not self.xml_object.xml_notations.xml_ornaments.get_children():
                self.xml_object.xml_notations.remove(self.xml_object.xml_notations.xml_ornaments)

            if self.xml_object.xml_notations.xml_dynamics and not self.xml_object.xml_notations.xml_dynamics.get_children():
                self.xml_object.xml_notations.remove(self.xml_object.xml_notations.xml_dynamics)

            if not self.xml_object.xml_notations.get_children():
                self.xml_object.remove(self.xml_object.xml_notations)

    def _update_xml_notehead(self):
        self.xml_object.xml_notehead = self.midi.notehead

    @property
    def is_tied(self) -> bool:
        """
        :return: True if an element :obj:`~musicxml.xmlelement.xmlelement.XMLTie` with type 'start' is under note's xml_object children.
        :rtype: bool
        """
        type_types = [t.type for t in self.xml_object.find_children('XMLTie')]
        if 'start' in type_types:
            return True
        else:
            return False

    @property
    def is_tied_to_next(self) -> bool:
        """
        :return: same as :obj:`~musicscore.musicscore.Note.is_tied`
        """
        return self.is_tied

    @property
    def is_tied_to_previous(self) -> bool:
        """
        :return: True if an element :obj:`~musicxml.xmlelement.xmlelement.XMLTie` with type 'stop' is under note's xml_object children.
        :rtype: bool
        """
        type_types = [t.type for t in self.find_children('XMLTie')]
        if 'stop' in type_types:
            return True
        else:
            return False

    @property
    def midi(self) -> Midi:
        """
        val must be a Midi object with a parent Chord. Midi with value 0 means rest.
        :return: note's :obj:`~musicscore.midi.Midi`.
        """
        return self._midi

    @midi.setter
    def midi(self, value):
        if not isinstance(value, Midi):
            raise TypeError('Note.midi property must be of type Midi')
        if not value.parent_chord:
            raise NoteMidiHasNoParentChordError
        self._midi = value
        self._midi.parent_note = self
        self._update_xml_pitch_or_rest()
        self._update_xml_accidental()

    @property
    def number_of_dots(self) -> int:
        """
        :return: number of dots
        :rtype: positive int
        """
        return self._number_of_dots

    @property
    def parent_chord(self) -> 'Chord':
        """
        :return: notes parent. Same as self.up
        :rtype: :obj:`musicscore.chord.Chord`
        """
        return self._parent_chord

    @QuarterDurationMixin.quarter_duration.setter
    def quarter_duration(self, value):
        if value is not None:
            if not self.parent_chord:
                raise NoteHasNoParentChordError()
            self._set_quarter_duration(value)
            self._update_xml_duration()
            self._update_xml_type()
        else:
            self.xml_object.xml_duration = None

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def get_children(self) -> List[Midi]:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.midi.Midi`]
        """
        return super().get_children()

    def get_parent(self) -> 'Chord':
        """
        :return: parent
        :rtype: :obj:`~musicscore.chord.Chord`
        """
        return super().get_parent()

    def get_parent_chord(self):
        return self.parent_chord

    def get_or_create_xml_notations(self) -> 'XMLNotations':
        """
        If note's ``xml_object`` has no :obj:`~musicxml.xmlelement.xmlelement.XMLNotations` as child this child will be created.

        :return: :obj:`~musicxml.xmlelement.xmlelement.XMLNotations`
        """
        if not self.xml_object.xml_notations:
            self.xml_object.xml_notations = XMLNotations()
        return self.xml_object.xml_notations

    def get_parent_measure(self) -> 'Measure':
        """
        :return: :obj:`~musicscore.measure.Measure` in note's ancestors
        """
        return self.parent_chord.get_parent_measure()

    def get_staff_number(self) -> int:
        """
        :return: number of :obj:`~musicscore.staff.Staff` in note's ancestors
        """
        midi_staff_number = self.midi.get_staff_number()
        if midi_staff_number:
            return midi_staff_number

        return self.parent_chord.get_staff_number()

    def get_voice_number(self) -> int:
        """
        :return: number of :obj:`~musicscore.voice.Voice` in note's ancestors
        """
        return self.get_parent_chord().get_voice_number()

    def remove_tie(self, type_: Optional[str] = None) -> None:
        """
        :param type_: 'start', 'stop', None: if None and note has :obj:`~musicxml.xmlelement.xmlelement.XMLTie` objects with both types
                      ValueError is raised.
        :return: None
        """
        ties = self.find_children('XMLTie')
        tie_to_be_removed = None
        if len(ties) == 0:
            pass
        elif len(ties) == 1:
            if type_ is None:
                tie_to_be_removed = ties[0]
            else:
                tie_to_be_removed = ties[0] if ties[0].type == type_ else None
        elif len(ties) == 2:
            if type_ is None:
                raise ValueError(
                    'Note has stop and start ties. Specify type_=start or type_=stop to decide which one should be removed')
            else:
                tie_to_be_removed = [t for t in ties if t.type == type_]
                tie_to_be_removed = None if not tie_to_be_removed else tie_to_be_removed[0]
        else:
            raise NotImplementedError
        if tie_to_be_removed:
            try:
                self.midi._ties.remove(tie_to_be_removed.type)
            except KeyError:
                pass
            tied_to_be_removed = \
                [t for t in self.xml_notations.find_children('XMLTied') if t.type == tie_to_be_removed.type][0]
            tie_to_be_removed.up.remove(tie_to_be_removed)
            xml_notations = tied_to_be_removed.up
            xml_notations.remove(tied_to_be_removed)
            if not xml_notations.get_children():
                xml_notations.up.remove(xml_notations)

    def start_tie(self) -> None:
        """
        Start a tie if not already started. Update xml_tie and xml_tied if necessary.
        """
        if not self.is_tied:
            self.xml_object.add_child(XMLTie(type='start'))
            self._set_xml_tied('start')
        self.midi._ties.add('start')

    def stop_tie(self) -> None:
        """
        Stop a tie if not already stopped. Update xml_tie and xml_tied if necessary.
        """
        if self.is_tied_to_previous:
            pass
        elif self.is_tied:
            self.find_children('XMLTie')[0].type = 'stop'
            self.xml_object.add_child(XMLTie(type='start'))
            self._set_xml_tied('stop')
        else:
            self.xml_object.add_child(XMLTie(type='stop'))
            self._set_xml_tied('stop')
        self.midi._ties.add('stop')

    def update_dots(self, number_of_dots: int) -> None:
        """
        Set or change number of dots

        :param number_of_dots: positiv int
        :return: None
        """
        self._number_of_dots = number_of_dots
        self._update_xml_dots(number_of_dots)

    def update_type(self, val: Optional[str] = None) -> None:
        """
        If val is None: type will be set according to note's quarter_duration.

        :param val: [‘1024th’, ‘512th’, ‘256th’, ‘128th’, ‘64th’, ‘32nd’, ‘16th’, ‘eighth’, ‘quarter’, ‘half’, ‘whole’, ‘breve’, ‘long’, ‘maxima’]
        :return: None
        """
        self._type = val
        if val is None:
            self._update_xml_type()
        else:
            self.xml_object.xml_type = val



from pathlib import Path

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.quarterduration import QuarterDuration
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloWorldQuantization(IdTestCase):
    def test_export_hello_world_quantization(self):
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a list of complicated quarter_durations
        """
        p = s.add_child(Part('p1'))
        import random
        random.seed(11)
        quarter_durations = []
        while sum(quarter_durations) < 40:
            quarter_durations.append(QuarterDuration(random.random() + random.randint(0, 3)))
        quarter_durations.append(44 - sum(quarter_durations))
        """
        He sets the possible subdivisions for all beats with quarter_duration and omits quintuplets and septuplets:
        """
        s.set_possible_subdivisions([2, 3, 4, 6, 8])
        """
        He adds chords with quarter_durations and writes each quarter_duration as lyrics
        """
        for qd in quarter_durations:
            ch = Chord(midis=71, quarter_duration=qd)
            ch.add_lyric(round(float(ch.quarter_duration), 2))
            p.add_chord(ch)
        """
        He sets part's get_quantized attribute to True.
        """
        p.quantize = True
        """
        # ... and exports the xml
        # """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



"""
Two voices share one staff. Each voice is assigned some lyrics.
"""

from pathlib import Path

from musicscore import Score, A, Chord, Lyrics, C, E, D, B, G
from musicscore.chord import Rest
from musicscore.tests.util import IdTestCase
from musicscore.util import slur_chords
from musicxml import XMLFermata, XMLAccent


class TestLily42a(IdTestCase):
    def test_lily_42a_MultiVoice_TwoVoicesOnStaff_Lyrics(self):
        score = Score()
        part = score.add_part('p1')
        v1 = [Chord(E(5), 2), Chord(D(5), 1), Chord(B(4), 1), Rest(1), Chord(D(5), 1), Chord(B(3), 1.5),
              Chord(C(5), 0.5), Rest(4)]
        v2 = [Chord(C(5), 2), Chord(B(4), 1), Chord(G(4), 1), Rest(1), Chord(B(4), 1), Chord(G(3), 1.5),
              Chord(A(4), 0.5), Rest(4)]
        for ch in [v1[0], v2[0]]:
            ch.add_x(XMLFermata())
            ch.add_x(XMLAccent())
        v1[1].add_dynamics('mf')
        slur_chords(v1[-3:-1])
        slur_chords(v2[-3:-1])
        [part.add_chord(ch, voice_number=1) for ch in v1]
        [part.add_chord(ch, voice_number=2) for ch in v2]
        Lyrics(['This', 'is', 'the', None, 'lyrics', 'of', 'Voice1', None], show_number=True,
               default_y=-110).add_to_chords(v1)
        Lyrics(['This', 'is', 'the', None, 'lyrics', 'of', 'Voice2', None], show_number=True,
               default_y=-130).add_to_chords(v2)
        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from typing import Union, Optional

from musicscore.util import isinstance_as_string
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLPageLayout, XMLPageMargins, XMLScaling, XMLDefaults, XMLSystemLayout, \
    XMLSystemMargins, XMLStaffLayout

__all__ = ['PAGE_MARGINS', 'PAGE_SIZES', 'SYSTEM_MARGINS', 'SYSTEM_LAYOUT', 'STAFF_LAYOUT', 'SCALING', 'Margins',
           'Scaling',
           'PageLayout', 'SystemLayout', 'StaffLayout']
#:
PAGE_MARGINS = {
    'A4': {
        'portrait': {'left': 140, 'right': 70, 'top': 70, 'bottom': 70},
        'landscape': {'left': 111, 'right': 70, 'top': 70, 'bottom': 70}
    }, 'A3': {
        'portrait': {'left': 111, 'right': 70, 'top': 70, 'bottom': 70},
        'landscape': {'left': 111, 'right': 70, 'top': 70, 'bottom': 70}
    }
}

#:
PAGE_SIZES = {'A4': (209.991, 297.0389), 'A3': (297.0389, 419.9819)}

#:
SYSTEM_MARGINS = {'left': 0, 'right': 0}

#:
SYSTEM_LAYOUT = {'system_distance': 117, 'top_system_distance': 117}

#:
STAFF_LAYOUT = {'staff_distance': 80}

#:
SCALING = {
    'millimeters': 7.2319,
    'tenths': 40
}


class LayoutMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent = None

    @property
    def parent(self):
        """
        :return: musicscore object which uses this layout object.

        .. todo::
           Print implementation: At this moment only :obj:`~musicscore.score.Score` is implemented.
        """
        return self._parent

    @parent.setter
    def parent(self, val):
        self._parent = val
        if isinstance(self, PageLayout):
            if isinstance_as_string(self.parent, 'Score'):
                self._update()
                self.parent.xml_object.xml_defaults.xml_page_layout = self.xml_object
            else:
                raise NotImplementedError
        elif isinstance_as_string(self, 'SystemLayout'):
            if isinstance_as_string(self.parent, 'Score'):
                self.parent.xml_object.xml_defaults.xml_system_layout = self.xml_object
            else:
                raise NotImplementedError
        elif isinstance_as_string(self, 'StaffLayout'):
            if isinstance_as_string(self.parent, 'Score'):
                self.parent.xml_object.xml_defaults.xml_staff_layout = self.xml_object
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError


class Margins:
    def __init__(self, parent: Union['PageLayout', 'SystemLayout'], left: Union[float, int] = None,
                 right: Union[float, int] = None,
                 top: Union[float, int] = None, bottom: Union[float, int] = None):
        self._parent = None
        self._left = None
        self._right = None
        self._top = None
        self._parent_xml_object = None
        self._parent = None

        self.parent = parent
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

    def _update_parent(self, side):
        if side not in ['left', 'right', 'top', 'bottom']:
            raise ValueError
        if side in ['top', 'bottom'] and isinstance(self.parent, SystemLayout):
            if eval(f"self.{side}") is not None:
                raise ValueError
        else:
            setattr(self._parent_xml_object, f"xml_{side}_margin", eval(f"self.{side}"))

    @property
    def bottom(self):
        """
        Sets and gets :obj:`~musicxml.xmlelement.xmlelement.XMLBottomMargin` value of parent.
        """
        return self._bottom

    @bottom.setter
    def bottom(self, val):
        self._bottom = val
        self._update_parent('bottom')

    @property
    def left(self) -> Union[int, float]:
        """
        Sets and gets :obj:`~musicxml.xmlelement.xmlelement.XMLLeftMargin` value of parent.
        """
        return self._left

    @left.setter
    def left(self, val):
        self._left = val
        self._update_parent('left')

    @property
    def parent(self):
        """
        :return: parent layout object. :obj:`PageLayout` and :obj:`SystemLayout` are implemented.
        """
        return self._parent

    @parent.setter
    def parent(self, val):
        self._parent = val
        if isinstance(val, PageLayout):
            self._parent_xml_object = self._parent.xml_object.xml_page_margins
        elif isinstance(val, SystemLayout):
            self._parent_xml_object = self._parent.xml_object.xml_system_margins
        else:
            raise NotImplementedError

    @property
    def right(self) -> Union[int, float]:
        """
        Sets and gets :obj:`~musicxml.xmlelement.xmlelement.XMLRightMargin` value of parent.
        """
        return self._right

    @right.setter
    def right(self, val):
        self._right = val
        self._update_parent('right')

    @property
    def top(self) -> Union[int, float]:
        """
        Sets and gets :obj:`~musicxml.xmlelement.xmlelement.XMLTopMargin` value of parent.
        """
        return self._top

    @top.setter
    def top(self, val) -> Union[int, float]:
        self._top = val
        self._update_parent('top')


class Scaling(XMLWrapper):
    _ATTRIBUTES = {'millimeters', 'tenths', 'score'}
    XMLClass = XMLScaling

    def __init__(self, millimeters: Union[int, float] = SCALING['millimeters'],
                 tenths: Union[int, float] = SCALING['tenths']):
        super().__init__()
        self._xml_object = self.XMLClass()
        self._millimeters = None
        self._tenths = None
        self._score = None

        self.millimeters = millimeters
        self.tenths = tenths

    def _update_score(self):
        if self.score:
            self.score.page_layout._set_page_height_and_width()

    @property
    def millimeters(self) -> Union[int, float]:
        """
        Sets and gets millimeters value of scaling object. After setting value, parent :obj:`~musicscore.score.Score`'s :obj:`PageLayout` is
        updated to reflect the changes.

        :return: millimeters
        :rtype: Union[int, float]
        """
        return self._millimeters

    @millimeters.setter
    def millimeters(self, val):
        if val != self._millimeters:
            self._millimeters = val
            self.xml_object.xml_millimeters = val
            self._update_score()

    @property
    def score(self):
        """
        Sets and gets parent :obj:`~musicscore.score.Score`. After setting score, its :obj:`~musicxml.xmlelement.xmlelement.XMLScaling`
        and :obj:`~musicxml.xmlelement.xmlelement.XMLDefaults` are created if needed.

        :return: parent score
        :rtype: :obj:`~musicscore.score.Score`
        """
        return self._score

    @score.setter
    def score(self, val):
        self._score = val
        if not self.score.xml_object.xml_defaults:
            self.score.xml_object.xml_defaults = XMLDefaults()
        self.score.xml_object.xml_defaults.xml_scaling = self.xml_object

    @property
    def tenths(self) -> Union[int, float]:
        """
        Sets and gets tenths value of scaling object. After setting value, parent :obj:`~musicscore.score.Score`'s :obj:`PageLayout` is updated to reflect the
        changes.

        :return: tenths
        :rtype: Union[int, float]
        """
        return self._tenths

    @tenths.setter
    def tenths(self, val):
        if val != self._tenths:
            self._tenths = val
            self.xml_object.xml_tenths = val
            self._update_score()

    def millimeters_to_tenths(self, x: Union[int, float]) -> Union[int, float]:
        """
        Converts millimeter value into tenths

        :param x: millimeters
        :return: calculated tenths
        """
        return round((x * self.tenths) / self.millimeters)


class PageLayout(XMLWrapper, LayoutMixin):
    """
    :param size: :obj:`PAGE_SIZES`
    :param orientation: 'portrait', 'landscape'
    """
    _ATTRIBUTES = {'scaling', 'size', 'orientation', 'parent'}
    XMLClass = XMLPageLayout

    def __init__(self, size: str = 'A4', orientation: str = 'portrait'):

        super().__init__()
        self._xml_object = self.XMLClass()
        self._xml_object.xml_page_margins = XMLPageMargins(type='both')

        self._size = None
        self._orientation = None

        self.size = size
        self.orientation = orientation

        self._margins = Margins(parent=self, **PAGE_MARGINS[self.size][self.orientation])

    def _get_page_height(self):
        return self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][1]) if self.orientation == 'portrait' else \
            self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][0])

    def _get_page_width(self):
        return self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][0]) if self.orientation == 'portrait' else \
            self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][1])

    def _set_page_height_and_width(self):
        self.xml_object.xml_page_height = self._get_page_height()
        self.xml_object.xml_page_width = self._get_page_width()

    def _update(self):
        self._set_page_height_and_width()
        self._margins = Margins(parent=self, **PAGE_MARGINS[self.size][self.orientation])

    @property
    def margins(self) -> Margins:
        """
        Gets margins attribute.

        :return: margins object
        :rtype: :obj:`Margins`
        """
        return self._margins

    @property
    def orientation(self) -> str:
        """
        Sets and gets orientation. Permitted values are ['portrait', 'landscape']. After setting value, if parent and size already
        exist, page's height and width are set.

        :return: 'portrait', 'landscape'
        :rtype: str
        """
        return self._orientation

    @orientation.setter
    def orientation(self, val):
        _permitted = ['portrait', 'landscape']
        if val not in _permitted:
            raise ValueError(f"{val} can only be: {_permitted}")
        if val != self._orientation:
            self._orientation = val
            if self.size and self.parent:
                self._update()

    @property
    def scaling(self) -> Scaling:
        """
        :return: :obj:`~musicscore.score.Score`'s :obj:`Scaling`.
        :rtype: :obj:`Scaling`
        """
        return self.parent.get_root().scaling

    @property
    def size(self) -> str:
        """
        Sets and gets orientation. Permitted values are keys of obj:`PAGE_SIZES`. After setting value, if parent and orientation already
        exist, page's height and width are set.

        :return: sizes in :obj:`PAGE_SIZES`
        :rtype: str
        """
        return self._size

    @size.setter
    def size(self, val):
        if val not in PAGE_SIZES:
            raise NotImplementedError
        if val != self._size:
            self._size = val
            if self.orientation and self.parent:
                self._update()


class SystemLayout(XMLWrapper, LayoutMixin):
    _ATTRIBUTES = {'system_distance', 'top_system_distance', 'parent'}
    XMLClass = XMLSystemLayout

    def __init__(self, system_distance: Union[int, float] = SYSTEM_LAYOUT['system_distance'],
                 top_system_distance: Union[int, float] = SYSTEM_LAYOUT['top_system_distance']):
        super().__init__()

        self._xml_object = self.XMLClass()
        self._xml_object.xml_system_margins = XMLSystemMargins()

        self.system_distance = system_distance
        self.top_system_distance = top_system_distance

        self._margins = Margins(parent=self, **SYSTEM_MARGINS)

    @property
    def margins(self) -> Margins:
        """
        Gets margins attribute.

        :return: margins object
        :rtype: :obj:`Margins`
        """
        return self._margins

    @property
    def system_distance(self) -> Optional[Union[int, float]]:
        """
        Sets and gets ``value`` of :obj:`~musicxml.xmlelement.xmlelement.XMLSystemDistance`.

        :return: ``self.xml_object.xml_system_distance.value_``
        :rtype: int, float, None
        """
        if self.xml_object.xml_system_distance:
            return self.xml_object.xml_system_distance.value_

    @system_distance.setter
    def system_distance(self, val):
        self.xml_object.xml_system_distance = val

    @property
    def top_system_distance(self) -> Optional[Union[int, float]]:
        """
        Sets and gets ``value`` of :obj:`~musicxml.xmlelement.xmlelement.XMLTopSystemDistance`.

        :return: ``self.xml_object.xml_top_system_distance.value_``
        :rtype: int, float, None
        """
        if self.xml_object.xml_top_system_distance:
            return self.xml_object.xml_top_system_distance.value_

    @top_system_distance.setter
    def top_system_distance(self, val):
        self.xml_object.xml_top_system_distance = val


class StaffLayout(XMLWrapper, LayoutMixin):
    _ATTRIBUTES = {'staff_distance', 'parent'}
    XMLClass = XMLStaffLayout

    def __init__(self, staff_distance=STAFF_LAYOUT['staff_distance']):
        super().__init__()
        self._xml_object = self.XMLClass()
        self.staff_distance = staff_distance

    @property
    def staff_distance(self) -> Optional[Union[int, float]]:
        """
        Sets and gets ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLStaffDistance`.

        :return: ``self.xml_object.xml_staff_distance.value_``
        :rtype: int, float, None
        """
        if self.xml_object.xml_staff_distance:
            return self.xml_object.xml_staff_distance.value_

    @staff_distance.setter
    def staff_distance(self, val):
        self.xml_object.xml_staff_distance = val



"""Beaming or slurs can indicate melismata for lyrics. Also make sure that notes without an explicit syllable are
treated as if they were part of a melisma."""

"""
Chord.break_beam() will be needed. Otherwise breaking the beams like in this test will effect all staves and voices
"""

from pathlib import Path

from musicscore import Score, Chord, A, Time
from musicscore.lyrics import Lyrics
from musicscore.tests.util import IdTestCase
from musicscore.util import slur_chords


class TestLily61h(IdTestCase):
    def test_lily_61h_BeamsMelismata(self):
        score = Score()
        part = score.add_part('p1')
        t = Time(4, 4)
        t.actual_signatures = (3, 8, 1, 8, 3, 8, 1, 8)
        part.add_measure(time=t)
        t = Time(4, 4)
        t.actual_signatures = 8 * [1, 8]
        part.add_measure(time=t)

        midi_pattern = [72, 69, 72, 69, 72, 67, 71, 0]
        chords = []
        for _ in range(4):
            chords += [Chord(m, 0.5) for m in midi_pattern]

        [part.add_chord(ch) for ch in chords]

        third_measure_chords = part.get_measure(measure_number=3).get_chords()
        slur_chords(third_measure_chords[:3])
        slur_chords(third_measure_chords[4:7])

        fourth_measure_chords = part.get_measure(measure_number=4).get_chords()
        slur_chords(fourth_measure_chords[:7])

        lyrics_pattern = [('Me', None, None, 'lis', 'ma', None, None)]
        for m in part.get_children():
            Lyrics(lyrics_pattern).add_to_chords(m.get_chords()[:-1])

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase


class TestChords(IdTestCase):
    def test_chords(self):
        s = Score()
        p = s.add_part('p1')
        midis = []
        for m in range(60, 72):
            midis = midis + [m]
            p.add_chord(Chord(midis=midis, quarter_duration=2))
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from musicscore.chord import Chord
from musicscore.dynamics import DYNAMICS, Dynamics
from musicxml.xmlelement.xmlelement import XMLPp, XMLFf, XMLSfpp, XMLP, XMLF

from musicscore.tests.util import ChordTestCase


class TestDynamics(ChordTestCase):
    def test_dynamics_init(self):
        d = Dynamics('pp')
        assert d.xml_object.__class__ == XMLPp

        d = Dynamics('sfpp')
        assert d.xml_object.__class__ == XMLSfpp

    def test_add_dynamics_to_a_chord_consequently(self):
        ch = Chord(60, 1)
        d1 = ch.add_dynamics('p')[0]
        d2 = ch.add_dynamics('f')[0]
        assert ch.xml_direction_types['below'] == [('dynamics', [d1]), ('dynamics', [d2])]
        ch._parent = self.mock_beat
        ch.finalize()
        assert len(ch._xml_directions) == 2
        assert isinstance(ch._xml_directions[0].xml_direction_type.xml_dynamics.get_children()[0], XMLP)
        assert isinstance(ch._xml_directions[1].xml_direction_type.xml_dynamics.get_children()[0], XMLF)

    def test_add_dynamics(self):
        for dynamics in DYNAMICS:
            ch = Chord(60, 1)
            ch._parent = self.mock_beat
            ch.add_dynamics(dynamics)
            ch.finalize()
            assert len(ch._xml_directions) == 1
            d = ch._xml_directions[0]
            assert d.placement == 'below'
            assert d.xml_direction_type.xml_dynamics.get_children()[0].__class__ == DYNAMICS[dynamics]

    def test_add_dynamics_placement(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        ch.add_dynamics('pp', placement='above')
        ch.finalize()
        d = ch._xml_directions[0]
        assert d.placement == 'above'

    def test_add_multiple_dynamics(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        ch.add_dynamics(['pp', 'ff'], 'below')
        ch.finalize()
        d = ch._xml_directions[0]
        dts = d.find_children('XMLDirectionType')
        assert d.placement == 'below'
        assert dts[0].xml_dynamics.get_children()[0].__class__ == XMLPp
        assert dts[1].xml_dynamics.get_children()[0].__class__ == XMLFf

        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        ch.add_dynamics('pp', 'above')
        ch.add_dynamics('ff', 'below')
        ch.finalize()
        ds = ch._xml_directions
        assert [ds[0].placement, ds[1].placement] == ['above', 'below']
        assert ds[0].xml_direction_type.xml_dynamics.get_children()[0].__class__ == XMLPp
        assert ds[1].xml_direction_type.xml_dynamics.get_children()[0].__class__ == XMLFf



import setuptools
from pathlib import Path

long_description = (Path(__file__).parent / "README.rst").read_text()
setuptools.setup(
    name="musicscore",
    version="2.0beta10",
    author="Alex Gorji",
    author_email="aligorji@hotmail.com",
    description="Generating musicxml files.",
    url="https://github.com/alexgorji/musicscore.git",
    packages=setuptools.find_packages(),
    install_requires=['musicxml==1.4', 'quicktions'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type='text/x-rst',
    include_package_data=True
)



from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase
from musicscore.util import XML_DIRECTION_TYPE_CLASSES
from musicxml.xmlelement.xmlelement import XMLSymbol, XMLDynamics, XMLWedge, XMLDashes, XMLBracket, XMLPedal, \
    XMLMetronome, XMLBeatUnit, XMLPerMinute, XMLOctaveShift, XMLHarpPedals, XMLPedalTuning, XMLPedalStep, XMLPedalAlter, \
    XMLStringMute, XMLScordatura, XMLAccord, XMLTuningOctave, XMLTuningStep, XMLImage, XMLPrincipalVoice, XMLPercussion, \
    XMLWood, XMLStaffDivide


class TestDynamics(IdTestCase):
    def test_direction_types(self):
        score = Score(title='Directions')
        p = score.add_part('part-1')
        for dt_class in [cl for cl in XML_DIRECTION_TYPE_CLASSES if cl != XMLDynamics]:
            if dt_class == XMLSymbol:
                dt_obj = dt_class('0')
            elif dt_class == XMLWedge:
                dt_obj = dt_class(type='crescendo')
            elif dt_class == XMLDashes:
                dt_obj = dt_class(type='start')
            elif dt_class == XMLBracket:
                dt_obj = dt_class(type='start', line_end='none')
            elif dt_class == XMLPedal:
                dt_obj = dt_class(type='start')
            elif dt_class == XMLMetronome:
                dt_obj = dt_class()
                dt_obj.add_child(XMLBeatUnit('quarter'))
                dt_obj.add_child(XMLPerMinute('120'))
            elif dt_class == XMLOctaveShift:
                dt_obj = dt_class(type='up')
            elif dt_class == XMLHarpPedals:
                dt_obj = dt_class()
                pt = dt_obj.add_child(XMLPedalTuning())
                pt.add_child(XMLPedalStep('A'))
                pt.add_child(XMLPedalAlter(1))
            elif dt_class == XMLStringMute:
                dt_obj = dt_class(type='on')
            elif dt_class == XMLScordatura:
                dt_obj = dt_class()
                acc = dt_obj.add_child(XMLAccord())
                acc.add_child(XMLTuningStep('A'))
                acc.add_child(XMLTuningOctave(0))
            elif dt_class == XMLImage:
                # dt_obj = dt_class(source='www.example.com')
                continue
            elif dt_class == XMLPrincipalVoice:
                dt_obj = dt_class(type='start', symbol='none')
            elif dt_class == XMLPercussion:
                dt_obj = dt_class()
                dt_obj.add_child(XMLWood('cabasa'))
            elif dt_class == XMLStaffDivide:
                dt_obj = dt_class(type='up')
            else:
                dt_obj = dt_class()

            chord = Chord(midis=60, quarter_duration=4)
            chord.add_direction_type(dt_obj)
            p.add_chord(chord)
        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



"""
Various time signatures: 2/2 (alla breve), 4/4 (C), 2/2, 3/2, 2/4, 3/4, 4/4, 5/4, 3/8, 6/8, 12/8
"""
from pathlib import Path

from musicscore import Score, generate_measures, Chord, C
from musicscore.tests.util import IdTestCase


class TestLily11a(IdTestCase):
    def test_lily_11a_TimeSignatures(self):
        score = Score('Time signatures')
        p = score.add_part('ts')

        for measure in generate_measures(
                [(2, 2), (4, 4), (2, 2), (3, 2), (2, 4), (3, 4), (4, 4), (5, 4), (3, 8), (6, 8), (12, 8)]):
            p.add_child(measure)
            p.add_chord(Chord(C(5), measure.quarter_duration))

        p.get_measure(1).time.symbol = 'cut'
        p.get_measure(2).time.symbol = 'common'

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



import inspect

from musicscore import Score, Part, Time, Chord, C
from musicscore.tests.util import IdTestCase, generate_repetitions, generate_path
from musicscore.util import chord_is_in_a_repetition


# MyXMLTestSuite test_cautionary

class TestShowAccidentalSigns(IdTestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part('part')
        self.measure_1 = self.part.add_measure()
        self.measure_2 = self.part.add_measure()
        self.staff_1_1 = self.measure_1.add_staff()
        self.staff_1_2 = self.measure_1.add_staff()
        self.staff_2_1 = self.measure_2.add_staff()
        self.staff_2_2 = self.measure_2.add_staff()
        self.voice_1_1_1 = self.staff_1_1.add_voice()
        self.voice_1_1_2 = self.staff_1_1.add_voice()
        self.voice_2_1_1 = self.staff_2_1.add_voice()
        self.voice_2_1_2 = self.staff_2_1.add_voice()
        all_names = ['score', 'part', 'measure_1', 'measure_2', 'staff_1_1', 'staff_1_2', 'staff_2_1', 'staff_2_2',
                     'voice_1_1_2', 'voice_2_1_1', 'voice_2_1_2']
        self.all = [eval(f"self.{name}", {'self': self}) for name in all_names]

    def reset_show_signs(self):
        for obj in self.all:
            obj.show_accidental_signs = None

    def test_show_accidental_signs_property(self):
        for obj in self.all:
            assert obj.show_accidental_signs == 'modern'
        self.score.show_accidental_signs = 'traditional'
        for obj in self.all:
            assert obj.show_accidental_signs == 'traditional'
        self.part.show_accidental_signs = 'modern'
        assert self.score.show_accidental_signs == 'traditional'
        for obj in self.all:
            if obj != self.score:
                assert obj.show_accidental_signs == 'modern'

        self.reset_show_signs()
        self.score.show_accidental_signs = 'traditional'
        for obj in self.all:
            assert obj.show_accidental_signs == 'traditional'

        self.measure_1.show_accidental_signs = 'modern'

        for obj in self.all:
            if obj in [self.score, self.part, self.measure_2, self.staff_2_1, self.staff_2_2, self.voice_2_1_1,
                       self.voice_2_1_2]:
                assert obj.show_accidental_signs == 'traditional'
            else:
                assert obj.show_accidental_signs == 'modern'

    def test_show_accidental_signs_repetition_signs_simple(self):
        s = Score()
        p = s.add_part('p1')
        p.add_measure(Time(3, 4))
        for qd in 3 * [1]:
            p.add_chord(Chord(C(5, '#'), qd))
        s.export_xml(generate_path(inspect.currentframe()))
        for index, chord in enumerate(p.get_chords()):
            if index == 0:
                assert chord_is_in_a_repetition(chord) is False
            else:
                assert chord_is_in_a_repetition(chord) is True

        for index, chord in enumerate(p.get_chords()):
            if index == 0:
                assert chord.midis[0].accidental.show is True
            else:
                assert chord.midis[0].accidental.show is False



from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLF, XMLFf, XMLFff, XMLFfff, XMLFffff, XMLFfffff, XMLFp, XMLFz, XMLMf, \
    XMLMp, XMLP, XMLPf, XMLPp, XMLPpp, XMLPppp, XMLPpppp, XMLPppppp, XMLRf, XMLRfz, XMLSf, XMLSffz, XMLSfp, XMLSfpp, \
    XMLSfz, XMLSfzp

__all__ = ['DYNAMICS', 'Dynamics']

#:
DYNAMICS = {"f": XMLF,
            "ff": XMLFf,
            "fff": XMLFff,
            "ffff": XMLFfff,
            "fffff": XMLFffff,
            "ffffff": XMLFfffff,
            "fp": XMLFp,
            "fz": XMLFz,
            "mf": XMLMf,
            "mp": XMLMp,
            "p": XMLP,
            "pf": XMLPf,
            "pp": XMLPp,
            "ppp": XMLPpp,
            "pppp": XMLPppp,
            "ppppp": XMLPpppp,
            "pppppp": XMLPppppp,
            "rf": XMLRf,
            "rfz": XMLRfz,
            "sf": XMLSf,
            "sffz": XMLSffz,
            "sfp": XMLSfp,
            "sfpp": XMLSfpp,
            "sfz": XMLSfz,
            "sfzp": XMLSfzp,
            }


class Dynamics(XMLWrapper):
    """
    This is a simple wrapper for dynamics objects. Value must be a key in :obj:`DYNAMICS`.
    """

    def __init__(self, value: str, *args, **kwargs):
        super().__init__()
        self.XMLClass = DYNAMICS[value]
        self._xml_object = self.XMLClass(*args, **kwargs)



import copy
from fractions import Fraction
from typing import Union, List, Optional, Any, Dict

from musicscore.clef import Clef
from musicscore.musictree import MusicTree
from musicscore.dynamics import Dynamics
from musicscore.exceptions import ChordAlreadySplitError, ChordCannotSplitError, ChordHasNoParentError, \
    ChordQuarterDurationAlreadySetError, AlreadyFinalizedError, DeepCopyException, ChordNotesAreAlreadyCreatedError, \
    ChordException, NotationException, ChordAddXException, ChordAddXPlacementException, RestCannotSetMidiError, \
    RestWithDisplayStepHasNoDisplayOctave, RestWithDisplayOctaveHasNoDisplayStep, GraceChordCannotHaveGraceNotes
from musicscore.finalize import FinalizeMixin
from musicscore.midi import Midi
from musicscore.note import Note
from musicscore.quarterduration import QuarterDuration, QuarterDurationMixin
from musicscore.util import XML_ARTICULATION_CLASSES, XML_TECHNICAL_CLASSES, XML_ORNAMENT_CLASSES, XML_DYNAMIC_CLASSES, \
    XML_OTHER_NOTATIONS, XML_DIRECTION_TYPE_CLASSES, XML_ORNAMENT_AND_OTHER_NOTATIONS, \
    XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS, isinstance_as_string
from musicxml.xmlelement.xmlelement import *

__all__ = ['Chord', 'Rest', 'GraceChord']

from musicxml.xmlelement.xmlelement import XMLElement

_all_articulations = Union[
    'XMLAccent', 'XMLStrongAccent', 'XMLStaccato', 'XMLTenuto', 'XMLDetachedLegato', 'XMLStaccatissimo',
    'XMLSpiccato', 'XMLScoop', 'XMLPlop', 'XMLDoit', 'XMLFalloff', 'XMLBreathMark', 'XMLCaesura', 'XMLStress',
    'XMLUnstress']

_all_technicals = Union[
    "XMLUpBow", "XMLDownBow", "XMLHarmonic", "XMLOpenString", "XMLThumbPosition", "XMLFingering", "XMLPluck", "XMLDoubleTongue",
    "XMLTripleTongue", "XMLStopped", "XMLSnapPizzicato", "XMLFret", "XMLString", "XMLHammerOn", "XMLPullOff", "XMLBend", "XMLTap",
    "XMLHeel", "XMLToe", "XMLFingernails", "XMLHole", "XMLArrow", "XMLHandbell", "XMLBrassBend", "XMLFlip", "XMLSmear", "XMLOpen",
    "XMLHalfMuted", "XMLHarmonMute", "XMLGolpe", "XMLOtherTechnical"]

_all_ornaments = Union[
    "XMLDelayedInvertedTurn", "XMLDelayedTurn", "XMLHaydn", "XMLInvertedMordent", "XMLInvertedTurn",
    "XMLInvertedVerticalTurn", "XMLMordent", "XMLOtherOrnament", "XMLSchleifer", "XMLShake", "XMLTremolo", "XMLTrillMark", "XMLTurn",
    "XMLVerticalTurn", "XMLWavyLine"
]

_all_dynamics = Union[
    "XMLF", "XMLFf", "XMLFff", "XMLFfff", "XMLFffff", "XMLFfffff", "XMLFp", "XMLFz", "XMLMf", "XMLMp", "XMLP", "XMLPf", "XMLPp", "XMLPpp", "XMLPppp",
    "XMLPpppp", "XMLPppppp", "XMLRf", "XMLRfz", "XMLSf", "XMLSffz", "XMLSfp", "XMLSfpp", "XMLSfz", "XMLSfzp"
]

_all_other_notations = Union[
    "XMLArpeggiate", "XMLFermata", "XMLFootnote", "XMLGlissando", "XMLLevel", "XMLNonArpeggiate", "XMLOtherNotation", "XMLSlide",
    "XMLSlur", "XMLAccidentalMark"
]


class Chord(MusicTree, QuarterDurationMixin, FinalizeMixin):
    """
    Chord is a sequence of one or more :obj:`~musicxml.xmlelement.xmlelement.XMLNote` s which occur at the same time in a :obj:`~musicxml.xmlelement.xmlelement.XMLMeasure` of a :obj:`~musicxml.xmlelement.xmlelement.XMLPart`.

    :param midis: :obj:`~musicscore.midi.Midi`, Midi.value, [Midi, Midi.value], 0 or [0] for a rest.
    :param quarter_duration: int, float, Fraction, :obj:`~musicscore.quarterduration.QuarterDuration` for duration counted in quarters (crotchets). 0 for grace note (or chord).
    """
    _ATTRIBUTES = {'midis', 'quarter_duration', 'notes', 'offset', 'split', 'voice', 'clef', 'metronome', 'arpeggio'}

    def __init__(self, midis: Optional[Union[List[Union[float, int]], List[Midi], float, int, Midi]] = None,
                 quarter_duration: Optional[Union[float, int, 'Fraction', QuarterDuration]] = None, **kwargs):
        self._midis = None
        self._xml_direction_types = {'above': [], 'below': []}

        self._xml_directions = []
        self._xml_lyrics = []
        self._xml_articulations = []
        self._xml_technicals = []
        self._xml_ornaments = []
        self._xml_dynamics = []
        self._xml_other_notations = []
        self._note_attributes = kwargs
        self._notes_are_set = False
        self._clef = None
        self._metronome = None
        self._grace_chords = {'before': [], 'after': []}
        self._arpeggio = None
        self._after_notes_xml_objects = []
        super().__init__(quarter_duration=quarter_duration)
        self._set_midis(midis)
        self.split = False
        self._original_starting_ties = None

    def _add_articulation(self, articulation, placement=None):
        if articulation.__class__ not in XML_ARTICULATION_CLASSES:
            raise ChordAddXException(f'{articulation} is not an articulation object.')
        if placement:
            try:
                articulation.placement = placement
            except AttributeError:
                raise ChordAddXPlacementException(f'{articulation} has to placement attribute.')
        self._xml_articulations.append(articulation)
        if self.notes:
            self._update_xml_articulations()

    def _add_child(self, child: Note) -> Note:
        """
        Check and add child to list of children. Child's parent is set to self.

        :param child: :obj:`~musicscore.note.Note`
        :return: child
        :rtype: :obj:`~musicscore.note.Note`
        """
        return super().add_child(child)

    def _add_direction_type(self, direction_type, placement=None):
        if direction_type.__class__ in XML_DYNAMIC_CLASSES:
            d = XMLDynamics()
            d.add_child(direction_type)
            direction_type = d
        if direction_type.__class__ not in XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS:
            raise ChordAddXException(f'{direction_type} is not a direction type object.')
        if placement:
            self.add_direction_type(direction_type, placement=placement)
        else:
            self.add_direction_type(direction_type)

    def _add_notation(self, notation, placement=None):
        if notation.__class__ in XML_DYNAMIC_CLASSES:
            d = XMLDynamics(placement=placement)
            d.add_child(notation)
            notation = d
        elif isinstance(notation, XMLFermata):
            if placement == 'above':
                notation.type = 'upright'
            elif placement == 'below':
                notation.type = 'inverted'
        elif notation.__class__ not in XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS:
            raise ChordAddXException(f'{notation} is not a notation type object.')
        elif placement:
            raise ChordAddXPlacementException(
                f'Chord.add_x({notation}) of parent_type notations cannot have a placement argument.')

        self._xml_other_notations.append(notation)
        if self.notes:
            self._update_xml_other_notations()

    def _add_ornament(self, ornament, placement=None):
        if ornament.__class__ not in XML_ORNAMENT_CLASSES + XML_ORNAMENT_AND_OTHER_NOTATIONS:
            raise ChordAddXException(f'{ornament} is not an ornament type object.')
        if placement:
            try:
                ornament.placement = placement
            except AttributeError:
                raise ChordAddXPlacementException(f'{ornament} has to placement attribute.')
        self._xml_ornaments.append(ornament)
        if self.notes:
            self._update_xml_ornaments()

    def _add_technical(self, technical, placement=None):
        if technical.__class__ not in XML_TECHNICAL_CLASSES:
            raise ChordAddXException(f'{technical} is not a technical object.')
        if placement:
            try:
                technical.placement = placement
            except AttributeError:
                raise ChordAddXPlacementException(f'{technical} has to placement attribute.')
        self._xml_technicals.append(technical)
        if self.notes:
            self._update_xml_technicals()

    def _set_original_starting_ties(self, original_chord):
        self._original_starting_ties = [copy.copy(midi._ties) for midi in original_chord.midis]

    def _set_midis(self, midis):
        if isinstance(midis, str):
            raise TypeError
        if hasattr(midis, '__iter__'):
            pass
        elif midis is None:
            midis = []
        else:
            midis = [midis]
        if len(midis) > 1 and 0 in midis:
            raise ValueError(
                'Chord cannot accept a mixed list of midis of rests and pitches or a list of more than one rests.')

        if 0 in midis and self.quarter_duration == 0:
            raise ValueError('A rest cannot be a grace note')
        self._midis = [Midi(v) if not isinstance(v, Midi) else v for v in midis]
        self._sort_midis()
        for midi in self._midis:
            midi._set_parent_chord(self)
        self._update_notes_pitch_or_rest()

    def _sort_midis(self):
        self._midis = sorted(self._midis)

    def _update_notes(self):
        if self._notes_are_set:
            raise ChordNotesAreAlreadyCreatedError()
        for index, midi in enumerate(self.midis):
            try:
                self.get_children()[index].midi = midi
            except IndexError:
                self._add_child(Note(midi=midi, **self._note_attributes))
        self._notes_are_set = True

    def _update_notes_pitch_or_rest(self):
        if self.notes:
            len_diff = len(self.notes) - len(self.midis)
            if len_diff > 0:
                to_be_removed = self.notes[len_diff:]
                for note in to_be_removed:
                    note.up.remove(note)
                    note.parent_chord = None
                    del note
            for index, m in enumerate(self.midis):
                if index < len(self.notes):
                    self.notes[index].midi = m
                else:
                    new_note = Note(midi=m, quarter_duration=self.quarter_duration)
                    self._add_child(new_note)

    def _update_notes_quarter_duration(self):
        for note in self.notes:
            note.quarter_duration = self.quarter_duration

    def _update_ties(self):
        # _update ties of already created notes
        for note in self.notes:
            note._update_ties()

    def _update_xml_notations_arpeggiate(self):
        if self.arpeggio:
            for n in self.notes:
                if not n.xml_notations:
                    n.xml_notations = XMLNotations()
                if self.arpeggio != 'none':
                    n.xml_notations.xml_arpeggiate = XMLArpeggiate()
                    if self.arpeggio != 'normal' and n == self.notes[0]:
                        n.xml_notations.xml_arpeggiate.direction = self.arpeggio
                else:
                    if n == self.notes[0]:
                        n.xml_notations.xml_non_arpeggiate = XMLNonArpeggiate(type='bottom')
                    elif n == self.notes[-1]:
                        n.xml_notations.xml_non_arpeggiate = XMLNonArpeggiate(type='top')

    def _update_xml_articulations(self):
        def _get_note_xml_articulations():
            try:
                return n.xml_notations.xml_articulations.get_children(ordered=False)
            except AttributeError:
                return []

        n = self.notes[0]

        note_articulations_not_in_chord = [art for art in _get_note_xml_articulations() if art not in
                                           self._xml_articulations]
        chord_articulations_not_in_note = [art for art in self._xml_articulations if
                                           art not in _get_note_xml_articulations()]

        if chord_articulations_not_in_note:
            n.get_or_create_xml_notations()
            if not n.xml_notations.xml_articulations:
                n.xml_notations.xml_articulations = XMLArticulations()

            for xml_articulation in chord_articulations_not_in_note:
                n.xml_notations.xml_articulations.add_child(xml_articulation)

        for art in note_articulations_not_in_chord:
            n.xml_notations.xml_articulations.remove(art)
        n._update_xml_notations()

    def _update_xml_directions(self):
        def _add_dynamics(list_of_dynamics, xml_direction):
            for dynamics in list_of_dynamics:
                dt = xml_direction.add_child(XMLDirectionType())
                dyn = dt.xml_dynamics = XMLDynamics()
                dyn.add_child(dynamics.xml_object)

        for placement in self._xml_direction_types:
            direction_types = self._xml_direction_types[placement]
            for direction_type in direction_types:
                d = XMLDirection(placement=placement)
                self._xml_directions.append(d)
                if hasattr(direction_type, '__iter__') and direction_type[0] == 'dynamics':
                    _add_dynamics(list_of_dynamics=direction_type[1], xml_direction=d)
                else:
                    dt = d.add_child(XMLDirectionType())
                    dt.add_child(direction_type)

    def _update_xml_dynamics(self):
        def _get_note_xml_dynamics():
            try:
                return n.xml_notations.xml_dynamics.get_children(ordered=False)
            except AttributeError:
                return []

        n = self.notes[0]

        note_dynamics_not_in_chord = [art for art in _get_note_xml_dynamics() if art not in
                                      self._xml_dynamics]
        chord_dynamics_not_in_note = [art for art in self._xml_dynamics if art not in _get_note_xml_dynamics()]

        if chord_dynamics_not_in_note:
            n.get_or_create_xml_notations()
            if not n.xml_notations.xml_dynamics:
                n.xml_notations.xml_dynamics = XMLDynamics(**self._xml_dynamics_kwargs)

            for xml_dynamic in chord_dynamics_not_in_note:
                n.xml_notations.xml_dynamics.add_child(xml_dynamic)

        for d in note_dynamics_not_in_chord:
            n.xml_notations.xml_dynamics.remove(d)

        n._update_xml_notations()

    def _update_xml_metronome(self):
        if self.metronome:
            d = XMLDirection(placement='above')
            self._xml_directions.append(d)
            dt = d.add_child(XMLDirectionType())
            dt.add_child(self.metronome.xml_object)
            if self.metronome.sound:
                d.add_child(self.metronome.sound)

    def _update_xml_ornaments(self):
        def _get_note_xml_ornaments():
            try:
                return n.xml_notations.xml_ornaments.get_children(ordered=False)
            except AttributeError:
                return []

        n = self.notes[0]

        note_ornaments_not_in_chord = [o for o in _get_note_xml_ornaments() if o not in
                                       self._xml_ornaments]
        chord_ornaments_not_in_note = [o for o in self._xml_ornaments if o not in _get_note_xml_ornaments()]

        if chord_ornaments_not_in_note:
            n.get_or_create_xml_notations()
            if not n.xml_notations.xml_ornaments:
                n.xml_notations.xml_ornaments = XMLOrnaments()

            for xml_ornament in chord_ornaments_not_in_note:
                n.xml_notations.xml_ornaments.add_child(xml_ornament)

        for o in note_ornaments_not_in_chord:
            n.xml_notations.xml_ornaments.remove(o)

        n._update_xml_notations()

    def _update_xml_other_notations(self):
        def _get_note_xml_other_notations():
            try:
                return [ch for ch in n.xml_notations.get_children(ordered=False) if ch.__class__ in XML_OTHER_NOTATIONS]
            except AttributeError:
                return []

        n = self.notes[0]

        note_other_notations_not_in_chord = [on for on in _get_note_xml_other_notations() if on not in
                                             self._xml_other_notations]
        chord_other_notations_not_in_note = [on for on in self._xml_other_notations if
                                             on not in _get_note_xml_other_notations()]

        if chord_other_notations_not_in_note:
            n.get_or_create_xml_notations()
            for xml_other_notation in chord_other_notations_not_in_note:
                n.xml_notations.add_child(xml_other_notation)

        for on in note_other_notations_not_in_chord:
            n.xml_notations.remove(on)

        n._update_xml_notations()

    def _update_xml_technicals(self):
        def get_note_xml_technical():
            try:
                return n.xml_notations.xml_technical.get_children(ordered=False)
            except AttributeError:
                return []

        n = self.notes[0]

        note_technicals_not_in_chord = [tech for tech in get_note_xml_technical() if tech not in
                                        self._xml_technicals]
        chord_technicals_not_in_note = [tech for tech in self._xml_technicals if tech not in get_note_xml_technical()]

        if chord_technicals_not_in_note:
            n.get_or_create_xml_notations()
            if not n.xml_notations.xml_technical:
                n.xml_notations.xml_technical = XMLTechnical()

            for xml_technical in chord_technicals_not_in_note:
                n.xml_notations.xml_technical.add_child(xml_technical)

        for tech in note_technicals_not_in_chord:
            n.xml_notations.xml_technical.remove(tech)

        n._update_xml_notations()

    def _update_xml_lyrics(self):

        n = self.notes[0]

        note_lyrics_not_in_chord = [lyric for lyric in n.xml_object.find_children('XMLLyric') if lyric not in
                                    self._xml_lyrics]
        chord_lyrics_not_in_note = [lyric for lyric in self._xml_lyrics if
                                    lyric not in n.xml_object.find_children('XMLLyric')]

        if chord_lyrics_not_in_note:

            for xml_lyric in chord_lyrics_not_in_note:
                n.xml_object.add_child(xml_lyric)

        for lyric in note_lyrics_not_in_chord:
            n.xml_object.remove(lyric)

    def _update_xml_chord(self):
        for n in self.notes[1:]:
            if not n.xml_object.xml_chord:
                n.xml_object.add_child(XMLChord())

    # public properties
    @property
    def all_midis_are_tied_to_next(self):
        if set([m.is_tied_to_next for m in self.midis]) == {True}:
            return True
        else:
            return False

    @property
    def all_midis_are_tied_to_previous(self):
        if set([m.is_tied_to_previous for m in self.midis]) == {True}:
            return True
        else:
            return False

    @property
    def arpeggio(self):
        return self._arpeggio

    @arpeggio.setter
    def arpeggio(self, val):
        permitted = [None, 'normal', 'up', 'down', 'none']
        if val not in permitted:
            raise ValueError(f'arpeggio value {val} must be in permitted list: {permitted}')
        if self._finalized:
            raise AlreadyFinalizedError(self, 'arpeggio.setter')
        self._arpeggio = val

    @property
    def clef(self):
        return self._clef

    @clef.setter
    def clef(self, val):
        if not isinstance(val, Clef):
            raise TypeError
        self._clef = val

    @property
    def is_rest(self) -> bool:
        """
        :return: ``True`` if Chord represents a rest, ``False`` if otherwise.
        :rtype: bool
        """
        if self._midis[0].value == 0:
            return True
        else:
            return False

    @property
    def is_tied_to_previous(self):
        return self.all_midis_are_tied_to_previous

    @property
    def is_tied_to_next(self):
        return self.all_midis_are_tied_to_next

    @property
    def metronome(self):
        return self._metronome

    @metronome.setter
    def metronome(self, val):
        if not isinstance_as_string(val, 'Metronome'):
            raise TypeError
        self._metronome = val

    @property
    def midis(self) -> List['Midi']:
        """
        :return: list of midis

        >>> ch = Chord(midis=60)
        >>> [type(m) for m in ch.midis]
        [<class 'musicscore.midi.Midi'>]
        >>> [m.value for m in ch.midis]
        [60]
        >>> ch = Chord(midis=[60, Midi(40)])
        >>> [m.value for m in ch.midis]
        [60, 40]
        >>> Chord([0, 60])
        Traceback (most recent call last):
        ...
        ValueError: Chord cannot accept a mixed list of midis of rests and pitches or a list of more than one rests.
        """
        return self._midis

    @midis.setter
    def midis(self, val):
        self._set_midis(val)

    @property
    def notes(self) -> List['Note']:
        """
        :return: :obj:`musicscore.chord.get_children` which are of type :obj:`musicscore.note.Note`.
        :rtype: List[:obj:`~musicscore.note.Note`]
        """
        return self.get_children()

    @property
    def offset(self) -> QuarterDuration:
        """
        :return: Offset in Chord's parent :obj:`~musicscore.beat.Beat`
        :rtype: QuarterDuration
        """
        if not self.up:
            return None
        elif self.previous is None:
            return 0
        else:
            return self.previous.offset + self.previous.quarter_duration

    @QuarterDurationMixin.quarter_duration.setter
    def quarter_duration(self, val):
        if self._quarter_duration is not None and self.up:
            raise ChordQuarterDurationAlreadySetError(
                'Chord is already attached to a Beat. Quarter Duration cannot be changed any more.')
        if val is not None:
            if self.midis and self.is_rest and val == 0:
                raise ValueError('A rest cannot be a grace note')
            self._set_quarter_duration(val)
            if self._notes_are_set:
                self._update_notes_quarter_duration()

    @property
    def voice_number(self) -> int:
        """
        :return: Number of parent :obj:`~musicscore.voice.Voice`
        :rtype: positive int
        """
        if not self.up:
            raise ChordHasNoParentError()
        if self.up and self.up.up:
            return self.up.up.number
        else:
            return 1

    @property
    def xml_articulations(self) -> List[_all_articulations]:
        """
        :return: list of xml articulations to be added to self.notes
        """
        return self._xml_articulations

    @property
    def xml_direction_types(self) -> Dict:
        """
        :return: dict of xml direction types to be added to self.notes.

        .. todo::
           Better documentation. Which types can be added?  Example?
        """
        return self._xml_direction_types

    @property
    def xml_lyrics(self) -> List['XMLLyric']:
        """
        :return: list of xml lyrics to be added to self.notes
        """
        return self._xml_lyrics

    @property
    def xml_technicals(self) -> List[_all_articulations]:
        """
        :return: list of xml technicals to be added to self.notes
        """
        return self._xml_technicals

    # public methods
    def add_after_note_xml_objects(self, xml_object):
        self._after_notes_xml_objects.append(xml_object)

    def add_direction_type(self, direction_type: XMLElement, placement: Optional[str] = None):
        if not placement:
            if isinstance(direction_type, XMLPedal) or isinstance(direction_type, XMLWedge):
                placement = 'below'
            else:
                placement = 'above'

        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_direction_type')
        if direction_type.__class__ not in XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS:
            raise TypeError(
                f'Wrong type {direction_type}. Possible classes: {XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS}')
        self._xml_direction_types[placement].append(direction_type)
        return direction_type

    def add_dynamics(self, dynamics: Union[List['Dynamics'], List['str'], 'Dynamics', str], placement: str = 'below') -> \
            List['Dynamics']:

        """
        This method is used to add one or more :obj:`musicscore.dynamics.Dynamics` objects to chord's private dictionary _xml_direction_types
        This list is used to create or _update directions of the first :obj:`~musicscore.note.Note` object of chord`s notes
        which are to be or are already created .

        :param dynamics: str, Dynamics of a list of Dynamics to be added to directions
        :param placement: above or below
        :return: List[:obj:`~musicscore.dynamics.Dynamics`]
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_dynamics')
        dynamics_list = [dynamics] if isinstance(dynamics, str) or not hasattr(dynamics, '__iter__') else list(
            dynamics)
        dynamics_object_list = [d if isinstance(d, Dynamics) else Dynamics(d) for d in dynamics_list]
        self._xml_direction_types[placement].append(('dynamics', dynamics_object_list))
        return dynamics_object_list

    def add_grace_chord(self, midis_or_grace_chord, type_=None, *, position=None):
        if self.up:
            raise ChordException(f'Chord {self} is already added to a measure. No grace chords can be added anymore.')
        if isinstance(midis_or_grace_chord, GraceChord):
            if type_:
                raise ValueError(f'Use GraceNote.type_ to set the type.')
            if position:
                raise ValueError(f'Use GraceNote.position to set the position.')
            gch = midis_or_grace_chord
        else:
            if not position:
                position = 'before'
            gch = GraceChord(midis_or_grace_chord, type_=type_, position=position)

        if gch.position == 'before':
            self._grace_chords['before'].append(gch)
        elif gch.position == 'after':
            self._grace_chords['after'].append(gch)
        gch.parent_chord = self
        return gch

    def add_tie(self, type_: str) -> None:
        """
        Chord's tie list is used to add ties to or _update ties of all midis and consequently :obj:`musicscore.note.Note`
        objects which are to be or are already created.

        :param type_: 'start' or 'stop'
        :return: None
        """

        for midi in self.midis:
            midi.add_tie(type_=type_)
        self._update_ties()

    def add_lyric(self, text: Union[Any, XMLLyric], **kwargs):
        """
        This method is used to add :obj:`~musicxml.xmlelement.xmlelement.XMLLyric` to chord's private _xml_lyricx list.
        This list is used to add lyrics to or _update lyrics of the first :obj:`~musicscore.note.Note` object of chord`s notes
        which are to be or are already created .

        :param text: if not of type :obj:`~musicxml.xmlelement.xmlelement.XMLLyric` a string conversion will be applied to text.
        :return: :obj:`~musicxml.xmlelement.xmlelement.XMLLyric`
        """
        if isinstance(text, XMLLyric):
            l = text
        else:
            l = XMLLyric(**kwargs)
            l.xml_text = str(text)
        self._xml_lyrics.append(l)
        if self.notes:
            self._update_xml_lyrics()
        return l

    def add_midi(self, midi):
        if self._notes_are_set:
            raise ChordNotesAreAlreadyCreatedError('Chord.add_midi cannot be used after creation of notes.')
        if not isinstance(midi, Midi):
            midi = Midi(midi)
        midi._set_parent_chord(self)
        self._midis.append(midi)
        self._sort_midis()
        return midi

    def add_wedge(self, wedge: Union['XMLWedge', str], placement: str = 'below') -> 'XMLWedge':
        """
        This method is used to add one :obj:`~musicxml.xmlelement.xmlelement.XMLWedge` object to chord's private
        dictionary _xml_direction_types This list is used to create or _update directions of the first
        :obj:`~musicscore.note.Note` object of chord`s notes which are to be or are already created .

        :param wedge: str, XMLWedge to be added to directions
        :param placement: above or below
        :return: :obj:`~musicxml.xmlelement.xmlelement.XMLWedge`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_wedge')
        wedge = XMLWedge(type=wedge) if isinstance(wedge, str) else wedge
        self.add_direction_type(wedge, placement=placement)
        return wedge

    def add_words(self, words: Union['XMLWords', str], placement: str = 'above', **kwargs) -> 'XMLWords':
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_words')

        if not isinstance(words, XMLWords):
            words = XMLWords(words, **kwargs)
        else:
            for key in kwargs:
                setattr(words, key, kwargs[key])
        return self.add_direction_type(words, placement=placement)

    def add_x(self, x: Union[_all_articulations, _all_technicals, _all_ornaments, _all_dynamics, _all_other_notations],
              *, placement=None, parent_type=None):
        """
        This method is used to add one xml object to a chord's private xml object lists (like _xml_articulations, xml_technicals
        etc.). These lists are used to add or _update articulations, technicals etc. of the first :obj:`~musicscore.note.Note` object of
        chord`s notes which are to be or are already created.

        :param x: musicxml articulation element, musicxml technical element, musicxml ornament element, musicxml dynamic element, musicxml other notations

        :return: x
        """
        if parent_type is None:
            if x.__class__ in XML_ARTICULATION_CLASSES:
                parent_type = 'articulation'
            elif x.__class__ in XML_TECHNICAL_CLASSES:
                parent_type = 'technical'
            elif x.__class__ in XML_ORNAMENT_CLASSES:
                parent_type = 'ornament'
            elif x.__class__ in XML_OTHER_NOTATIONS:
                parent_type = 'notation'
            elif x.__class__ in XML_DIRECTION_TYPE_CLASSES:
                parent_type = 'direction_type'
            elif x.__class__ in XML_ORNAMENT_AND_OTHER_NOTATIONS:
                permitted_parent_types = ['notation', 'ornament']
                raise NotationException(f'{x} is ambivalent. Set parent type {permitted_parent_types}.')
            elif x.__class__ in XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS or x.__class__ in XML_DYNAMIC_CLASSES:
                permitted_parent_types = ['notations', 'direction_type']
                raise NotationException(f'{x} is ambivalent. Set parent type {permitted_parent_types}.')
            else:
                raise ValueError(f'parent_type of {x} could not be determined.')

        if parent_type == 'articulation':
            self._add_articulation(x, placement=placement)
        elif parent_type == 'technical':
            self._add_technical(x, placement=placement)
        elif parent_type == 'ornament':
            self._add_ornament(x, placement=placement)
        elif parent_type == 'notation':
            self._add_notation(x, placement=placement)
        elif parent_type == 'direction_type':
            self._add_direction_type(x, placement=placement)
        else:
            raise NotImplementedError(f'parent_type: {parent_type} not implemented.')

        return x

    def finalize(self):
        """
        Finalize can be called only once. All necessary updates and xmlelement object creations will take place and the MusicTree
        object will be prepared for returning a musicxml snippet or a whole musicxml file.

        - Check if parent :obj:`~musicscore.beat.Beat` exists.
        - Ancestor :obj:`~musicscore.measure.Measure._update_divisions()` is called to _update :obj:`~musicxml.xmlelement.xmlelement.XMLMeasure`'s :obj:`~musicxml.xmlelement.xmlelement.XMLDivisions` attribute.
        - Following updates are triggered: update_notes, update_xml_chord, update_notes_quarter_durations, update_xml_lyrics,
          update_xml_directions, update_xml_articulations, update_technicals
        """
        if self._finalized:
            raise AlreadyFinalizedError(self)

        if not self.up:
            raise ChordHasNoParentError('Chord needs a parent Beat to create notes.')

        self._update_notes()
        self._update_xml_chord()

        self._update_notes_quarter_duration()
        self._update_xml_lyrics()
        self._update_ties()
        self._update_xml_directions()
        self._update_xml_metronome()
        self._update_xml_articulations()
        self._update_xml_technicals()
        self._update_xml_ornaments()
        self._update_xml_dynamics()
        self._update_xml_other_notations()
        self._update_xml_notations_arpeggiate()
        self._finalized = True

    def has_same_pitches(self, other: 'Chord') -> bool:
        """
        Only for chords with pitches. Rest chords cannot use this method.

        :param other: Other chord to which the comparison takes place
        :return: `True` if pitches of self and other chord has the same pitch parameters and accidental values else `False`
        """
        if not isinstance(other, Chord):
            raise TypeError
        if self.is_rest or other.is_rest:
            raise TypeError('Rest cannot use method has_same_pitches.')
        if [m.value for m in self.midis] != [m.value for m in other.midis]:
            return False
        for m1, m2 in zip(self.midis, other.midis):
            #     if m1.accidental.show != m2.accidental.show:
            #         return False
            if m1.accidental.get_pitch_parameters() != m2.accidental.get_pitch_parameters():
                return False
        return True

    def get_brackets(self):
        return self.get_x(XMLBracket)

    def get_children(self) -> List[Note]:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.note.Note`]
        """
        return super().get_children()

    def get_grace_chords(self, position='before'):
        return self._grace_chords[position]

    def get_x(self, type_):
        if type_ == XMLDynamics:
            raise NotImplementedError(f'get_x of type_ {type_} not Implemented.')
        elif type_ in XML_DIRECTION_TYPE_CLASSES:
            output = []
            output += [x for x in self._xml_direction_types['above'] if isinstance(x, type_)]
            output += [x for x in self._xml_direction_types['below'] if isinstance(x, type_)]
            return output
        elif type_ in XML_ORNAMENT_CLASSES:
            return [x for x in self._xml_ornaments if isinstance(x, type_)]
        elif type_ in XML_TECHNICAL_CLASSES:
            return [x for x in self._xml_technicals if isinstance(x, type_)]
        elif type_ in XML_ARTICULATION_CLASSES:
            return [x for x in self._xml_articulations if isinstance(x, type_)]
        elif type_ in XML_OTHER_NOTATIONS:
            return [x for x in self._xml_other_notations if isinstance(x, type_)]
        elif type_ in XML_ORNAMENT_AND_OTHER_NOTATIONS:
            return [x for x in self._xml_ornaments if isinstance(x, type_)] + [x for x in
                                                                               self._xml_other_notations if
                                                                               isinstance(x, type_)]
        else:
            raise NotImplementedError(f'get_x of type_ {type_} not Implemented.')

    def get_parent(self) -> 'Beat':
        """
        :return: parent
        :rtype: :obj:`~musicscore.beat.Beat`
        """
        return super().get_parent()

    def get_parent_measure(self) -> 'Measure':
        """
        :return: parent measure
        """
        return self.up.up.up.up

    def get_slurs(self):
        return self.get_x(XMLSlur)

    def get_staff_number(self):
        try:
            return self.up.up.up.number
        except AttributeError:
            return None

    def get_voice(self):
        raise TypeError

    def get_voice_number(self) -> int:
        """
        :return: parent voice number
        :rtype: positive int
        """
        try:
            return self.up.up.number
        except AttributeError:
            return None

    def get_wedges(self):
        return self.get_x(XMLWedge)

    def get_words(self):
        return self.get_x(XMLWords)

    def set_possible_subdivisions(self):
        raise TypeError

    def split_and_add_beatwise(self, beats: List['Beat']) -> List['Chord']:
        """
        All betas must have a voice parent

        :param beats:
        :return:
        """
        voice_set = {beat.up for beat in beats}
        if len(voice_set) != 1:
            raise ChordCannotSplitError('Beats have must have a single Voice as common ancestor.')

        voice = voice_set.pop()
        if voice is None:
            raise ChordCannotSplitError('Beats have no parent.')

        if voice.get_children()[
           voice.get_children().index(beats[0]): voice.get_children().index(beats[-1]) + 1] != beats:
            raise ChordCannotSplitError("Beats as Voice's children has another order as input list of beats")

        if beats[0] != voice.get_current_beat():
            raise ChordAlreadySplitError('First beat must be the next beat in voice which can accept chords.')
        if beats[-1] != voice.get_children()[-1]:
            raise ChordAlreadySplitError('Last beat must be the last beat in voice.')
        quarter_durations = self.quarter_duration.get_beatwise_sections(
            offset=beats[0].filled_quarter_duration, beats=beats)
        self.quarter_duration = quarter_durations[0][0]
        self.split = True

        if self._original_starting_ties is None:
            self._set_original_starting_ties(self)
        voice.get_current_beat().add_child(self)
        current_chord = self
        output = [self]
        for qd in quarter_durations[0][1:]:
            copied = _split_copy(self, qd)
            copied.split = True
            voice.get_current_beat().add_child(copied)

            current_chord.add_tie('start')
            copied.add_tie('stop')
            for midi in copied.midis:
                midi.accidental.show = False
            current_chord = copied
            output.append(current_chord)
        if quarter_durations[1]:
            # left over
            leftover_chord = _split_copy(self, quarter_durations[1])
            current_chord.add_tie('start')
            leftover_chord.add_tie('stop')
            for midi in leftover_chord.midis:
                midi.accidental.show = False
        else:
            leftover_chord = None
        self.up.up.leftover_chord = leftover_chord
        if not leftover_chord and output[-1]._original_starting_ties:
            for ties, midi in zip(output[-1]._original_starting_ties, output[-1].midis):
                if 'start' in ties:
                    midi.add_tie('start')
        if leftover_chord and leftover_chord._original_starting_ties:
            for ties, midi in zip(leftover_chord._original_starting_ties, leftover_chord.midis):
                if 'start' in ties:
                    midi.add_tie('start')
        return output

    def to_rest(self) -> None:
        """
        Set self.midis to [0]

        :return: None
        """
        self.midis = [0]

    def __setattr__(self, key, value):
        if key[0] != '_' and key not in self._ATTRIBUTES.union(self._TREE_ATTRIBUTES) and key not in self.__dict__:
            if self.notes:
                if isinstance(value, str) or not hasattr(value, '__iter__'):
                    value = [value] * len(self.notes)
                for n, v in zip(self.notes, value):
                    setattr(n, key, v)
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item):
        if not self._notes_are_set:
            raise AttributeError(f"AttributeError: 'Chord' object has no attribute '{item}'")
        output = [getattr(n, item) for n in self.notes]
        if output and callable(output[0]):
            raise AttributeError(f"Chord cannot call Note method {item}. Call this method on each note separately")
        return output

    def __deepcopy__(self, memodict={}):
        '''
        Only midi and quarter_duration are deepcopied. _ties are copied.
        Not included in deepcopy at the moment:
        self._xml_direction_types
        self._xml_directions = []
        self._xml_lyrics = []
        self._xml_articulations = []
        self._xml_technicals = []
        self._xml_ornaments = []
        self._xml_dynamics = []
        self._xml_articulations_kwargs = {}
        self._xml_technicals_kwargs = {}
        self._xml_ornaments_kwargs = {}
        self._xml_dynamics_kwargs = {}
        self._xml_other_notations = []
        self._note_attributes = kwargs
        self.split
        '''
        if self._notes_are_set:
            raise DeepCopyException("After setting notes, Midi cannot be deepcopied anymore. ")
        copied = self.__class__(midis=[midi.__deepcopy__() for midi in self.midis],
                                quarter_duration=self.quarter_duration.__deepcopy__())
        return copied


class GraceChord(Chord):
    _ATTRIBUTES = Chord._ATTRIBUTES.union({'type_', 'parent_chord', 'position'})

    def __init__(self, midis: Optional[Union[List[Union[float, int]], List[Midi], float, int, Midi]] = None, *,
                 type_=None, position='before', **kwargs):
        super().__init__(midis=midis, quarter_duration=0, **kwargs)
        self._type = None
        self._position = None
        self.type_ = type_
        self.position = position

    @Chord.quarter_duration.setter
    def quarter_duration(self, val):
        if val != 0:
            raise ChordException('quarter_duration of a GraceChord is always 0 and cannot be set')
        else:
            self._quarter_duration = 0

    def _update_xml_type(self):
        for n in self.notes:
            n.xml_object.xml_type = self.type_

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        permitted = ['before', 'after']
        if val not in permitted:
            raise ValueError(f'Wrong position. Permitted are: {permitted}')
        else:
            self._position = val

    @property
    def type_(self):
        return self._type

    @type_.setter
    def type_(self, val):
        if val is None:
            self._type = None
        else:
            if not isinstance(val, XMLType):
                self._type = XMLType(val)
            else:
                self._type = val

    def add_grace_chord(self, midis_or_grace_chord, type_=None, *, position=None):
        raise GraceChordCannotHaveGraceNotes

    def get_grace_chords(self, position='before'):
        raise GraceChordCannotHaveGraceNotes

    def finalize(self):
        super().finalize()
        self._update_xml_type()


class Rest(Chord):
    _ATTRIBUTES = Chord._ATTRIBUTES.union({'display_step', 'display_octave', 'measure'})

    def __init__(self, quarter_duration, display_step=None, display_octave=None, measure=None, **kwargs):
        if 'midis' in kwargs.keys():
            raise RestCannotSetMidiError
        super().__init__(midis=0, quarter_duration=quarter_duration, **kwargs)
        self._display_step = None
        self._display_octave = None
        self.display_step = display_step
        self.display_octave = display_octave
        self._measure = None
        self.measure = measure

    @property
    def display_step(self):
        return self._display_step

    @display_step.setter
    def display_step(self, val):
        permitted = [None, 'A', 'B', 'C', 'D', 'E', 'F', 'G']
        if val not in permitted:
            raise TypeError(f'display_step value {val} not in permitted list {permitted}')
        self._display_step = val

    @property
    def display_octave(self):
        return self._display_octave

    @display_octave.setter
    def display_octave(self, val):
        if val and (not isinstance(val, int) or val < 1):
            raise TypeError(f'display_octave value {val} can only be None or a positive integer. ')
        self._display_octave = val

    @property
    def measure(self):
        return self._measure

    @measure.setter
    def measure(self, val):
        self._measure = val

    def finalize(self):
        if self.display_step and not self.display_octave:
            raise RestWithDisplayStepHasNoDisplayOctave({self.display_step})

        if self.display_octave and not self.display_step:
            raise RestWithDisplayOctaveHasNoDisplayStep({self.display_octave})

        super().finalize()
        if self.display_step and self.display_octave:
            self.notes[0].xml_rest.xml_display_step = self.display_step
            self.notes[0].xml_rest.xml_display_octave = self.display_octave
        if self.measure:
            self.notes[0].xml_rest.measure = self.measure


def _split_copy(chord: Chord, new_quarter_duration: Union[QuarterDuration, Fraction, int, float] = None) -> Chord:
    """
    This function is used when a chord needs to be split. It creates a copy of the chord with a new quarter_duration object. All midis
    will be deepcopied. No attributes like lyrics, articulations etc. will be added to the copy.

    :param chord:
    :param new_quarter_duration: quarter_duration of copied chord. If None a shallow copy of chord.quarter_duration is created.
    :return: copied chord
    """
    if new_quarter_duration is None:
        new_quarter_duration = chord.quarter_duration.__copy__()
    new_chord = Chord(midis=[m._copy_for_split() for m in chord.midis], quarter_duration=new_quarter_duration)
    new_chord._original_starting_ties = chord._original_starting_ties
    return new_chord


def _group_chords(chords: List[Chord], quarter_durations: List[Union[QuarterDuration, Fraction, int, float]]) -> \
        Optional[List[List[Chord]]]:
    """
    A creates a nested list of chords. Chords can be divided into groups. Each group has its own specific quarter duration sum.

    :param chords:
    :param quarter_durations:
    :return: Optional[List[List[:obj:`Chord`]]]
    """
    if sum(c.quarter_duration for c in chords) != sum(quarter_durations):
        raise ValueError
    output = []
    for _ in quarter_durations:
        output.append([])
    index = 0
    current_quarter_duration = quarter_durations[0]
    for ch in chords:
        output[index].append(ch)
        current_sum = sum(c.quarter_duration for c in output[index])
        if current_sum < current_quarter_duration:
            pass
        elif current_sum == current_quarter_duration:
            index += 1
            if index == len(quarter_durations):
                pass
            else:
                current_quarter_duration = quarter_durations[index]
        else:
            return None
    return output



from pathlib import Path

from musicscore.chord import Chord
from musicscore.measure import Measure
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloPianoStaves(IdTestCase):
    def test_export_hello_world_piano_staves(self):
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a part
        """
        p = s.add_child(Part('P1', name='Music'))
        """
        He adds some chords for two staves with each two voices
        """
        for qd in 4 * [1]:
            p.add_chord(Chord(60, qd), voice_number=3, staff_number=1)
            p.add_chord(Chord(60 - 24, qd), voice_number=3, staff_number=2)
        for qd in 8 * [1 / 2]:
            p.add_chord(Chord(72, qd), voice_number=2, staff_number=1)
            p.add_chord(Chord(72 - 24, qd), voice_number=2, staff_number=2)
        for qd in 12 * [1 / 3]:
            p.add_chord(Chord(84, qd), voice_number=1, staff_number=1)
            p.add_chord(Chord(84 - 24, qd), voice_number=1, staff_number=2)
        p.add_measure()
        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from typing import Optional, Union, List

from musicscore.beat import Beat
from musicscore.chord import GraceChord
from musicscore.exceptions import VoiceHasNoBeatsError, VoiceHasNoParentError, VoiceIsFullError, \
    AddChordError, AlreadyFinalizedError
from musicscore.musictree import MusicTree
from musicscore.finalize import FinalizeMixin
from musicscore.quantize import QuantizeMixin
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLVoice

__all__ = ['Voice']


class Voice(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    _ATTRIBUTES = {'number', 'leftover_chord', 'is_filled'}
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)
    XMLClass = XMLVoice

    def __init__(self, number=None, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(value_='1', *args, **kwargs)
        self._number = None
        self.number = number
        self._current_beat_index = None
        self._leftover_chord = None
        self._final_updated = False

    def _add_chord(self, chord: 'Chord') -> List['Chord']:
        """
        :param chord: :obj:`~musicscore.chord.Chord`, required
        :return: added chord or a list of split chords
        """
        if not self.get_children():
            raise VoiceHasNoBeatsError

        try:
            current_beat = self.get_children()[self.get_current_beat_index()]
        except IndexError:
            raise VoiceIsFullError(f'Voice number {self.value_} of Measure number {self.up.up.number} is full.')

        if isinstance(chord, GraceChord) and chord.position == 'after':
            return current_beat.add_child(chord)

        if current_beat.is_filled:
            self._current_beat_index += 1
            return self._add_chord(chord)
        else:
            return current_beat.add_child(chord)

    @property
    def is_filled(self) -> bool:
        """
        :return: ``True`` if voice has :obj:`~musicscore.beat.Beat` children and the last child is filled, else ``False``
        """
        if self.get_children():
            return self.get_children()[-1].is_filled
        else:
            return False

    @property
    def leftover_chord(self) -> Optional['Chord']:
        """
        :return: None or a :obj:`~musicscore.chord.Chord` which is left over after adding a chord to the voice.
        """
        return self._leftover_chord

    @leftover_chord.setter
    def leftover_chord(self, val):
        self._leftover_chord = val

    @property
    def number(self) -> Optional[int]:
        """
        :type: None or int. If None number is set to 1.
        :return: positive int or None
        """
        if self._number is None:
            return None
        object_value = self.xml_object.value_
        if object_value is not None:
            return int(object_value)

    @number.setter
    def number(self, val):
        self._number = val
        if val is not None:
            self.xml_object.value_ = str(val)
        else:
            self.xml_object.value_ = '1'

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def add_beat(self, beat_quarter_duration: Optional[Union['QuarterDuration', 'Fraction', int, float]] = 1) -> Beat:
        """
        Creates and adds a :obj:`~musicscore.beat.Beat` to voice

        :param beat_quarter_duration: if None beat_quarter_duration is set to 1.
        :return: :obj:`~musicscore.beat.Beat`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_beat')
        if beat_quarter_duration is None:
            beat_quarter_duration = 1
        return self.add_child(Beat(beat_quarter_duration))

    def add_chord(self, *args, **kwargs):
        raise AddChordError()

    def add_child(self, child: Beat) -> Beat:
        """
        Check and add child to list of children. Child's parent is set to self.

        :param child: :obj:`~musicscore.beat.Beat`
        :return: child
        :rtype: :obj:`~musicscore.beat.Beat`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_child')
        if not self.up:
            raise VoiceHasNoParentError('A child Beat can only be added to a Voice if voice has a Staff parent.')
        return super().add_child(child)

    def get_children(self) -> List[Beat]:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.beat.Beat`]
        """
        return super().get_children()

    def get_parent(self) -> 'Staff':
        """
        :return: parent
        :rtype: :obj:`~musicscore.staff.Staff`
        """
        return super().get_parent()

    def get_current_beat_index(self) -> int:
        """
        :return: First not filled child :obj:`~musicscore.beat.Beat`
        """
        if not self.get_children():
            raise ValueError('Voice has no beats.')
        else:
            if not self._current_beat_index:
                self._current_beat_index = 0
        return self._current_beat_index

    def get_current_beat(self) -> 'Beat':
        try:
            current_beat = self.get_children()[self.get_current_beat_index()]
        except IndexError:
            raise VoiceIsFullError()
        if current_beat.is_filled:
            self._current_beat_index += 1
            return self.get_current_beat()
        return current_beat

    def update_beats(self, *quarter_durations) -> Optional[List[Beat]]:
        """
        Creates and adds or replaces Beats.

        :param quarter_durations: if None and a measure as ancestor exists, this measure's
                                  :obj:`musicscore.time.Time.get_beats_quarter_durations()` method is called.
        :return: None if quarter_durations is None and no measures as ancestor exists, else list of created beats.
        """
        if not quarter_durations:
            if self.up and self.up.up:
                quarter_durations = self.up.up.time.get_beats_quarter_durations()
            else:
                return
        else:
            if len(quarter_durations) == 1 and hasattr(quarter_durations[0], '__iter__'):
                quarter_durations = quarter_durations[0]

        self.remove_children()

        for quarter_duration in quarter_durations:
            self.add_child(Beat(quarter_duration))
        return self.get_children()
